/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2016-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/datastructures/skeletoninfo.h>
#include "inviwo/core/util/volumeramutils.h"
#include "../util/skeletonadjacencymapper.h"

namespace inviwo {

namespace skeletonutil {

#pragma optimize( "", off )

SkeletonInfo::SkeletonInfo(std::shared_ptr<const Volume> maskVolume,
                                               std::shared_ptr<const Volume> distVolume)
    : index_(size3_t(maskVolume->getDimensions()))
    , dim_(size3_t(maskVolume->getDimensions()))
    , maskVolume_(maskVolume)
    , distVolume_(distVolume)
    , thres_(0.0f) {
    ivec3 rdim(ivec3(dim_) - ivec3(1));

    // transform voxel index from distance volume to closest matching index in the binary mask
    auto distVolDim = dvec3(distVolume_->getDimensions() - size3_t(1));
    auto maskVolDim = dvec3(maskVolume_->getDimensions() - size3_t(1));

    isForeGround_ = [&, distVolDim, maskVolDim](const ivec3& index) {
        const auto maskIndex = size3_t(dvec3(index) * maskVolDim / distVolDim);
        if (glm::any(glm::greaterThanEqual(maskIndex, maskVolume_->getDimensions()))) {
            LogWarnCustom("SkeletonInfo", "source index " << index << ", dest = " << maskIndex);
        }
        return (maskVolume_->getRepresentation<VolumeRAM>()->getAsNormalizedDouble(maskIndex) > 0.0);
    };

    voxelBoundaryTest_ = [&](ivec3 newPos) {
        if (newPos.x < 0 || newPos.y < 0 || newPos.z < 0 || newPos.x > rdim.x ||
            newPos.y > rdim.y || newPos.z > rdim.z) {
            return false;
        }
        return true;
    };

    voxelMapping_ = [&](ivec3 newPos) { return index_(newPos); };

    voxelgetNormalizedPos_ = [&](ivec3 newPos) { return vec3(newPos) / vec3(rdim); };

    voxelNeighbourHood_ = [&](ivec3 pos, double radius) {
        std::vector<size3_t> neighbours;

        auto nodes = tree_.findCloseTo(pos, radius);
        for (auto& tnode : nodes) {
            auto p = tnode->get().pos_;
            auto indxp = voxelMapping_(p);
            // std::cout << "Getting Neighbourhood around " << pos << " radius:" << radius <<
            // std::endl;
            if (allVoxels_.find(indxp) != allVoxels_.end() &&
                allVoxels_.find(indxp)->second.hasVoxelFlag(VoxelFlags::ForeGround)) {
                neighbours.push_back(p);
            }
        }
        return neighbours;
    };

    voxelNeighbourHoodVoxels_ = [&](ivec3 pos, double radius) {

        std::vector<Voxel> neighbours;
        auto nodes = tree_.findCloseTo(pos, radius);
        for (auto& tnode : nodes) {
            auto p = tnode->get();
            auto indxp = p.index_;
            // std::cout << "Getting Neighbourhood around " << pos << " radius:" << radius <<
            // std::endl;
            if (allVoxels_.find(indxp) != allVoxels_.end() &&
                allVoxels_.find(indxp)->second.hasVoxelFlag(VoxelFlags::ForeGround)) {
                auto it = foregroundVoxels_.find(indxp);
                if (it->second.hasVoxelFlag(VoxelFlags::ForeGround))
                    neighbours.push_back(it->second);
            }
        }
        return neighbours;
    };
}



void SkeletonInfo::fillupVoxelInfo() {

    std::function<void(const size3_t&)> func = [&](const size3_t& pos) {};

    // construct voxels
    func = [&](const size3_t& pos) {

        auto cpos = glm::clamp(ivec3(pos), ivec3(0), ivec3(dim_) - ivec3(1));
        auto& tree = tree_;

        // Voxel flags: exterior, queued, deletable

        ivwAssert(allVoxels_.find(index_(cpos)) == allVoxels_.end(), "Voxel entry already exist");

        // allVoxels[index(cpos)] = Voxel();
        // Voxel& v = allVoxels[index(cpos)];
        Voxel v;
        v.init();
        v.pos_ = cpos;
        v.index_ = index_(cpos);
        v.distance_ = (distVolume_->getRepresentation<VolumeRAM>()->getAsNormalizedDouble(pos));
        v.value_ = (maskVolume_->getRepresentation<VolumeRAM>()->getAsNormalizedDouble(pos));
        v.neighbours_ = 0;
        v.flags_ = 0;
        v.labels_ = 0;
        v.segmentationValue_ = 0.0;
        v.mapping_ = voxelMapping_;
        v.boundaryTest_ = voxelBoundaryTest_;
        v.getNeighbours_ = voxelNeighbourHood_;
        v.getNeighbourVoxels_ = voxelNeighbourHoodVoxels_;
        v.getNormalizedVoxelPos_ = voxelgetNormalizedPos_;

        for (int x = 0; x < 3; x++)
            for (int y = 0; y < 3; y++)
                for (int z = 0; z < 3; z++) v.unsetNeighbour(ivec3(x, y, z));

        v.setVoxelFlag(VoxelFlags::None);

        if (isForeGround_(cpos))
            v.setVoxelFlag(VoxelFlags::ForeGround);
        else
            v.setVoxelFlag(VoxelFlags::BackGround);

        allVoxels_[index_(cpos)] = v;

        tree_.insert(cpos, v);

    };
    util::forEachVoxel(*maskVolume_->getRepresentation<VolumeRAM>(), func);
}

#pragma optimize( "", on )

}  // namespace

}  // namespace
