/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2016-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/datastructures/skeletonvoxels.h>
#include <inviwo/core/util/indexmapper.h>
#include <modules/morphologyvisualization/util/skeletonadjacencymapper.h>

namespace inviwo {

const util::IndexMapper3D skeletonutil::Voxel::mapper3d_ = util::IndexMapper3D(size3_t(3));


namespace skeletonutil {

void Voxel::init() {
	pos_ = vec3(-1.0);
	index_ = 0;
	distance_ = -1.0f;
	value_ = 0.0f;
	neighbours_ = 0;
	flags_ = VoxelFlags::None;
	segmentationValue_ = 0.0;
	nPos_ = pos_;

	labels_ = 0;
	mapping_ = [&](size3_t newPos) {
		return true;
	};
	boundaryTest_ = [&](size3_t newPos) {
		return true;
	};
	getNeighbours_ = [&](size3_t newPos, double radius=2.0) {
		std::vector<size3_t> defaultList = {newPos};
		return defaultList;
	};
	getNeighbourVoxels_ = [&](size3_t newPos, double radius = 2.0) {
		std::vector<Voxel> defaultList;
		return defaultList;
	};
}

/*
Voxel::Voxel(const Voxel& rhs) {
	copy(rhs);
}

Voxel& Voxel::operator=(const Voxel& that) {
	return *this;
}

void Voxel::copy(const Voxel& rhs) {
	pos_ = rhs.pos_;
	index_ = rhs.index_;
	distance_ = rhs.distance_;
	value_ = rhs.value_;
	neighbours_ = rhs.neighbours_;
	flags_ = rhs.flags_;
	segmentationValue_ = rhs.segmentationValue_;
	nPos_ = rhs.nPos_;
	//mapper3d_ = rhs.mapper3d_;

	mapping_ = rhs.mapping_;
	boundaryTest_ = rhs.boundaryTest_;
	getNeighbours_ = rhs.getNeighbours_;
	getNeighbourVoxels_ = rhs.getNeighbourVoxels_;
}
*/

bool isVoxelMaximalCenter(const Voxel& v) {
	auto sRadius = (v.distance_);

	int allNeighbourCount = 0;

	NAdjacencyMapper nm;

	auto neigbourIndcies = v.getNeighbourVoxels_(v.pos_, 1.5);
	for (auto& ngh : neigbourIndcies) {
		if (ngh.index_ == v.index_) {
			continue;
		}
		auto nRadius = (ngh.distance_);
		if (sRadius < nRadius)
			allNeighbourCount++;
	}

	if (!allNeighbourCount) return true;
	return false;
}

bool isVoxelMinimalCenter(const Voxel& v) {
	auto sRadius = (v.distance_);

	int allNeighbourCount = 0;
	NAdjacencyMapper nm;

	auto neigbourIndcies = v.getNeighbourVoxels_(v.pos_, 1.5);
	for (auto& ngh : neigbourIndcies) {
		if (ngh.index_ == v.index_) {
			continue;
		}
		auto nRadius = (ngh.distance_);
		if (sRadius > nRadius)
			allNeighbourCount++;
	}

	if (!allNeighbourCount) return true;
	return false;
}

std::vector<Voxel> getMaximalCenters(const VoxelMap& sourceVoxels,
                                     double distanceThreshold) {

	std::vector<Voxel> maximas;

	if (sourceVoxels.size() == 0) return maximas;

	for (auto& sv : sourceVoxels) {
		if (sv.second.distance_ <= distanceThreshold) continue;
		if (isVoxelMaximalCenter(sv.second) ||
			isVoxelMinimalCenter(sv.second))
			maximas.push_back(sv.second);
	}

	return maximas;
}

std::vector<Voxel> getSkeletonCenters(VoxelMap& sourceVoxels,
	double distanceThreshold, int itr) {

	std::vector<Voxel> maximas;

	if (sourceVoxels.size() == 0) return maximas;

	VoxelMap sourceVoxelCopy = sourceVoxels;
	
	double runningDistanceThreshold = distanceThreshold;

	for (int i = 0; i < itr; i++) {

		dvec2 currThresMinMax(0.0);
		for (auto& sv : sourceVoxels) {
			if (sv.second.distance_ > runningDistanceThreshold) {
				currThresMinMax.x = glm::min(currThresMinMax.x, runningDistanceThreshold);
				currThresMinMax.y = glm::max(currThresMinMax.y, runningDistanceThreshold);
			}
			else sv.second.distance_ = 0;
		}

		for (auto& sv : sourceVoxels) {
			if (sv.second.distance_ > runningDistanceThreshold) {
				if (isVoxelMaximalCenter(sv.second) ||
					isVoxelMinimalCenter(sv.second)) {
					maximas.push_back(sv.second);
					sv.second.distance_ = 0;
				}
			}
		}

		//runningDistanceThreshold = currThresMinMax.x;
	}
	sourceVoxels = sourceVoxelCopy;
	return maximas;
}

std::unordered_map<size_t, std::vector<Voxel>> getLabeledVoxelSegments(const VoxelMap& sourceVoxels,
                                                                       double distanceThreshold) {
    VoxelMap input;
	input.reserve(sourceVoxels.size() + 1);

    for (auto& sv : sourceVoxels) {
        if (sv.second.distance_ >= distanceThreshold) {
			input[sv.second.index_] = sv.second;
			input[sv.second.index_].labels_ = 1;
        }
    }

    std::unordered_map<size_t, std::vector<Voxel>> segments;


	VoxelMap output;
	for (auto& mv : input) {
		output[mv.first] = mv.second;
		output[mv.first].labels_ = 0;
	}

    NAdjacencyMapper nm;

    auto neighbourOffsets = nm.getN26(ivec3(0));

    auto labelNo = 2;
    for (auto& mv : input) {

		if (output[mv.first].labels_!=0) continue;

        int usedLabels = 0;
		
		//nm.LabelRecusively(input, labelNo, mv.second.pos_, neighbourOffsets, output, usedLabels, lastVisited);
		nm.LabelSerially(input, labelNo, mv.second.pos_, neighbourOffsets, output, usedLabels, 1);
		if (usedLabels>0) 
			labelNo++;	
    }

	size_t segmentCount = 1;
    for (size_t i = 2; i < labelNo; i++) {
        auto currentLabel = i;

        std::vector<Voxel> segmentedVoxels;
        for (auto& o : output) {
            if (o.second.labels_ == currentLabel) {
                segmentedVoxels.push_back(o.second);
            }
        }
        
        if (segmentedVoxels.size())
            segments[segmentCount++] = segmentedVoxels;
        
    }

    return segments;
}

}

}  // namespace
