/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2016-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#ifndef IVW_SKELETONVOXELS_H
#define IVW_SKELETONVOXELS_H

#include <modules/morphologyvisualization/morphologyvisualizationmoduledefine.h>
#include <inviwo/core/common/inviwo.h>
#include <inviwo/core/util/indexmapper.h>

namespace inviwo {

namespace skeletonutil {

enum VoxelFlags {
    None = 0,
    ForeGround = 1 << 0,
    BackGround = 1 << 1,
    Interior = 1 << 2,
    Exterior = 1 << 3,
    Skeleton = 1 << 4,
    Maximum = 1 << 5,
    MaximumInterface = 1 << 6,
    Deletable = 1 << 7,
    UnDeletable = 1 << 8,
    Queued = 1 << 9,
    EndPoint = 1 << 10,
};

struct IVW_MODULE_MORPHOLOGYVISUALIZATION_API Voxel {

    vec3 pos_; //!< 3D voxel index
    size_t index_;  //!< linear voxel index
    double value_;  
    double distance_;   //!< value from distance transform
    double segmentationValue_;
    size_t neighbours_;
    size_t flags_;
    vec3 nPos_; //nearest background voxel
    static const util::IndexMapper3D mapper3d_;
    
    //these are callbacks used during any voxel operations
    int labels_;
    std::function<size_t(size3_t)> mapping_;
    std::function<bool(size3_t)> boundaryTest_;
    std::function<vec3(size3_t)> getNormalizedVoxelPos_;
    std::function<std::vector<size3_t>(size3_t, double radius)> getNeighbours_;
    std::function<std::vector<Voxel>(size3_t, double radius)> getNeighbourVoxels_;

    Voxel() {
        init();
    }

    //copy cons
    /*Voxel(const Voxel& rhs);
    Voxel& operator=(const Voxel& that);
    void copy(const Voxel& rhs);*/

    // index
    void init();
    void setVoxelFlag(VoxelFlags flag) { flags_ |= flag; }
    void unsetVoxelFlag(VoxelFlags flag) { flags_ &= ~flag; }
    bool hasVoxelFlag(VoxelFlags flag) const { return (flags_ & flag) == flag; }

    void setNeighbour(int flag) { neighbours_ |= flag; }
    void unsetNeighbour(int flag) { neighbours_ &= ~flag; }
    bool hasNeighbour(int flag) const noexcept { return (neighbours_ & flag) == flag; }

    // size3 index
    void setNeighbour(size3_t cellid) {
        auto nflag = 1 << (int)mapper3d_(cellid);
        // std::cout << "Add Exterior Info :" << cellid << "Flags :" << nflag << std::endl;
        setNeighbour(nflag);
    }

    void unsetNeighbour(size3_t cellid) {
        auto nflag = 1 << (int)mapper3d_(cellid);
        unsetNeighbour(nflag);
        return;
    }

    bool hasNeighbour(size3_t cellid) const {
        auto nflag = 1 << (int)mapper3d_(cellid);
        return hasNeighbour(nflag);
    }
};

typedef std::unordered_map<size_t, Voxel> VoxelMap;

std::vector<Voxel> IVW_MODULE_MORPHOLOGYVISUALIZATION_API getMaximalCenters(const VoxelMap& sourceVoxels, double distanceThreshold=0.8);

std::vector<Voxel> IVW_MODULE_MORPHOLOGYVISUALIZATION_API getSkeletonCenters(VoxelMap& sourceVoxels, double distanceThreshold, int itr=1);

std::unordered_map<size_t, std::vector<Voxel>> IVW_MODULE_MORPHOLOGYVISUALIZATION_API getLabeledVoxelSegments(const VoxelMap& sourceVoxels, double distanceThreshold = 0.8);

bool IVW_MODULE_MORPHOLOGYVISUALIZATION_API isVoxelMaximalCenter(const Voxel& v);
bool IVW_MODULE_MORPHOLOGYVISUALIZATION_API isVoxelMinimalCenter(const Voxel& v);

}

}  // namespace

#endif  // IVW_SKELETONVOXELS_H
