/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2017-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/processors/chargepools.h>

#include <modules/gromacs/processors/gromacsutil.h>
#include <inviwo/core/datastructures/geometry/mesh.h>
#include <inviwo/core/util/clock.h>
#include <inviwo/core/datastructures/buffer/bufferram.h>
#include <inviwo/core/datastructures/buffer/bufferramprecision.h>
#include <inviwo/core/datastructures/geometry/basicmesh.h>
#include <inviwo/core/datastructures/volume/volumeram.h>
#include <inviwo/core/datastructures/volume/volumeramprecision.h>
#include <inviwo/core/util/indexmapper.h>
#include <inviwo/core/datastructures/transferfunction.h>
#include <modules/opengl/texture/textureunit.h>
#include <modules/opengl/volume/volumegl.h>

#include <modules/morphologyvisualization/util/gradientsampler.h>

#include <algorithm>
#include <numeric>
#include <array>

//#pragma optimize("", off)

namespace inviwo {

// The Class Identifier has to be globally unique. Use a reverse DNS naming scheme
const ProcessorInfo ChargePools::processorInfo_{
    "org.inviwo.ChargePools",   // Class identifier
    "Charge Pools",             // Display name
    "MorphologyVisualization",  // Category
    CodeState::Experimental,    // Code state
    Tags::CPU,                  // Tags
};
const ProcessorInfo ChargePools::getProcessorInfo() const { return processorInfo_; }

ChargePools::ChargePools()
    : Processor()
    , gromacsInport_("gromacsIn")
    , distanceInport1_("distance1")
    , maximaVolOutport_("maximaOut")
    , meshOutport_("meshOut")
    , colorMaxima_("colorMaxima", "Color Maxima", vec4(0.4f, 0.4f, 0.4f, 1.0f), vec4(0.0f),
                   vec4(1.0f))
    , colorMaximaInterface_("colorMaximaInterface", "Color Interface Maxima",
                            vec4(0.8f, 0.8f, 0.8f, 1.0f), vec4(0.0f), vec4(1.0f)) {
    addPort(gromacsInport_);
    addPort(distanceInport1_);
    addPort(maximaVolOutport_);
    addPort(meshOutport_);

    colorMaxima_.setSemantics(PropertySemantics::Color);
    colorMaximaInterface_.setSemantics(PropertySemantics::Color);

    addProperty(colorMaxima_);
    addProperty(colorMaximaInterface_);
}

void ChargePools::process() {
    {
        IVW_CPU_PROFILING_IF(150, "maxima computation");

        auto result = findMaxima(distanceInport1_.getData());

        maximaVolOutport_.setData(result.first);
        meshOutport_.setData(getMaximaMesh(result.second, result.first->getDimensions()));
    }
}

std::pair<std::shared_ptr<Volume>, std::vector<ivec3>> ChargePools::findMaxima(
    const std::shared_ptr<const Volume>& distVolume) {
    if (distVolume->getDataFormat() != DataFormat<float>::get()) {
        throw Exception("ChargePools processor only supports float distance volumes");
    }

    auto volume =
        static_cast<const VolumeRAMPrecision<float>*>(distVolume->getRepresentation<VolumeRAM>());

    auto data = volume->getDataTyped();

    auto dstRepresentation = std::make_shared<VolumeRAMPrecision<glm::i16>>(volume->getDimensions());
    auto dstVolume = std::make_shared<Volume>(dstRepresentation);
    auto dstData = dstRepresentation->getDataTyped();

    std::fill(dstData, dstData + glm::compMul(dstVolume->getDimensions()), 0);

    const ivec3 dim(volume->getDimensions());
    util::IndexMapper<3, int> im(dim);

    const std::array<ivec3, 26> offsets = {{
        {-1, -1, -1}, {0, -1, -1}, {1, -1, -1}, {-1, 0, -1}, {0, 0, -1}, {1, 0, -1}, {-1, 1, -1},
        {0, 1, -1},   {1, 1, -1},  {-1, -1, 0}, {0, -1, 0},  {1, -1, 0}, {-1, 0, 0}, {1, 0, 0},
        {-1, 1, 0},   {0, 1, 0},   {1, 1, 0},   {-1, -1, 1}, {0, -1, 1}, {1, -1, 1}, {-1, 0, 1},
        {0, 0, 1},    {1, 0, 1},   {-1, 1, 1},  {0, 1, 1},   {1, 1, 1},
    }};

    auto indexValid = [dim](const ivec3& pos) {
        return (glm::all(glm::greaterThanEqual(pos, ivec3(0))) &&
                glm::all(glm::lessThan(pos, ivec3(dim))));
    };
    auto makeValid = [dim](const ivec3& pos) {
        return ivec3(ivec2(pos + dim) % ivec2(dim), pos.z);
    };

    auto lexicographicalComp = [](const ivec3& index1, const ivec3& index2) {
        return (index1.x < index2.x) || ((index1.y < index2.y) && (index1.x <= index2.x)) ||
               ((index1.z < index2.z) && (index1.x <= index2.x) && (index1.y <= index2.y));
    };

    std::vector<ivec3> maxima;

    std::function<int(const ivec3&)> walker = [&](const ivec3& index) -> int {
        if (dstData[im(index)] != 0) {
            return dstData[im(index)];
        }

        float currentValue = data[im(index)];
        auto currentIndex(index);

        if (currentValue == 0.0f) {
            return 0;
        }

        for (auto offset : offsets) {
            // skip indices outside volume
            const auto newPos(makeValid(index + offset));

            if (!indexValid(newPos)) {
                continue;
            }
            if (data[im(newPos)] != currentValue) {
                if (data[im(newPos)] > data[im(currentIndex)]) {
                    currentValue = data[im(newPos)];
                    currentIndex = newPos;
                }
            } else {
                if (lexicographicalComp(newPos, currentIndex)) {
                    currentIndex = newPos;
                }
            }
        }
        if (currentIndex == index) {
            // maxima found, assign ID
            maxima.push_back(index);
            const auto id = static_cast<int>(maxima.size());
            dstData[im(index)] = static_cast<glm::i16>(id);
            return id;
        } else {
            // not a maxima, check neighbor recursively
            auto id = walker(currentIndex);
            dstData[im(index)] = static_cast<glm::i16>(id);;
            return id;
        }
    };

    for (int z = 0; z < dim.z - 1; ++z) {
        for (int y = 0; y < dim.y; ++y) {
            for (int x = 0; x < dim.x; ++x) {
                walker(ivec3(x, y, z));
            }
        }
    }

    dstVolume->dataMap_.dataRange = dvec2(0.0, maxima.size());
    dstVolume->dataMap_.valueRange = dvec2(0.0, maxima.size());

    // HOTFIX: enable nearest neighbor sampling for volume
    TextureUnit unit;
    dstVolume->getRepresentation<VolumeGL>()->bindTexture(unit.getEnum());
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    dstVolume->getRepresentation<VolumeGL>()->unbindTexture();

    return {dstVolume, maxima};
}

std::shared_ptr<Mesh> ChargePools::getMaximaMesh(const std::vector<ivec3>& maximaIndices,
                                                 const size3_t& volDim) {
    if (maximaIndices.empty()) {
        return std::make_shared<BasicMesh>();
    }

    std::shared_ptr<BasicMesh> mesh = std::make_shared<BasicMesh>();

    TransferFunction tf((int)512);
    tf.clearPoints();
    tf.addPoint(vec2(0.0f, 1.0f), vec4(1.0f, 0.0f, 0.0f, 1.0f));  // red
    tf.addPoint(vec2(0.5f, 1.0f), vec4(1.0f, 1.0f, 1.0f, 1.0f));  // white
    tf.addPoint(vec2(1.0f, 1.0f), vec4(0.0f, 0.0f, 1.0f, 1.0f));  // blue

    const vec4 colorMax3D(colorMaxima_.get());
    const vec4 colorMaxSurface(colorMaximaInterface_.get());

    auto isInterface = [&, dim = ivec3(volDim) ](const ivec3& index) {
        return ((index.z == 0) || (index.z == dim.z - 1));
    };

    std::vector<BasicMesh::Vertex> vertices;
    vertices.reserve(maximaIndices.size());
    for (auto& elem : maximaIndices) {
        // map value from distance transform to color
        // auto color(tf.sample(elem.second.distance_));

        /*
        if (elem.second.hasVoxelFlag(VoxelFlags::Maximum)) {
        ...
        }
        */

        auto color = isInterface(elem) ? colorMaxSurface : colorMax3D;
        // create
        vertices.push_back({elem, vec3(0.0f, 1.0f, 0.0f), elem, color});
    }

    mesh->addVertices(vertices);

    auto ind = mesh->addIndexBuffer(DrawType::Points, ConnectivityType::None);
    std::vector<std::uint32_t> indices;
    indices.resize(maximaIndices.size());
    std::iota(indices.begin(), indices.end(), 0);
    ind->append(indices);

    // update mesh matrices
    glm::mat4 modelMatrix, worldMatrix;
    gromacsutil::getModelAndWorldMatricesForGromacsData(*gromacsInport_.getData(), modelMatrix,
                                                        worldMatrix);

    mesh->setModelMatrix(modelMatrix);
    mesh->setWorldMatrix(worldMatrix);

    return mesh;
}

//#pragma optimize("", on)

}  // namespace inviwo
