/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2017-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/processors/linestotube.h>
#include <modules/morphologyvisualization/util/trajectoryspringmass.h>

namespace inviwo {

// The Class Identifier has to be globally unique. Use a reverse DNS naming scheme
const ProcessorInfo LinesToTube::processorInfo_{
    "org.inviwo.LinesToTube",  // Class identifier
    "Lines To Tube",           // Display name
    "Morphology",              // Category
    CodeState::Experimental,   // Code state
    Tags::CPU,                 // Tags
};
const ProcessorInfo LinesToTube::getProcessorInfo() const { return processorInfo_; }

LinesToTube::LinesToTube()
    : Processor()
    , meshIn_("meshIn")
    , meshOut_("meshOut")
    , glyphsize_("glyphsize", "Glyph Size", 1.0f, 0.01f, 100.0f)
    , glyphScale_("glyphScale", "Glyph Scale", 1.0f, 0.01f, 10.0f)
    , showHeadSegments_("showHeadSegments", "Show Head Segments", true)
    , headSegment_("headArrowSegment", "Head Segment", 0.0f, 0.15f, 0.0f, 1.0f, 0.001f, 0.001f)
    , showBodySegments_("showBodySegments", "Show Body Segments", true)
    , bodySegment_("bodySegment", "Body Segment", 0.45f, 0.65f, 0.0f, 1.0f, 0.001f, 0.001f)
    , showTailSegments_("showTailSegments", "Show Tail Segments", true)
    , tailSegment_("tailArrowSegment", "Tail Segment", 0.85f, 1.0f, 0.0f, 1.0f, 0.001f, 0.001f) {
    addPort(meshIn_);
    addPort(meshOut_);
    addProperty(glyphsize_);
    addProperty(glyphScale_);

    addProperty(showHeadSegments_);
    addProperty(headSegment_);
    addProperty(showBodySegments_);
    addProperty(bodySegment_);
    addProperty(showTailSegments_);
    addProperty(tailSegment_);

    showHeadSegments_.onChange([&] { headSegment_.setVisible(showHeadSegments_); });

    showBodySegments_.onChange([&] { bodySegment_.setVisible(showBodySegments_); });

    showTailSegments_.onChange([&] { tailSegment_.setVisible(showTailSegments_); });
}

void LinesToTube::process() {
    auto clipHead = headSegment_.get();
    auto clipBody = bodySegment_.get();
    auto clipTail = tailSegment_.get();

    auto clipRanges = std::vector<vec2>();

    auto mesh = std::make_shared<BasicMesh>();

    if (showHeadSegments_) {
        clipRanges.push_back({clipHead.x, clipHead.y});
    }

    if (showBodySegments_) {
        clipRanges.push_back({clipBody.x, clipBody.y});
    }

    if (showTailSegments_) {
        clipRanges.push_back({clipTail.x, clipTail.y});
    }

    auto glyphOption = (trajectoryutils::LineTubeGlyphOption::AllCylinder |
                        trajectoryutils::LineTubeGlyphOption::ArrowInTheEnd);

    if (clipRanges.size()) {
        mesh = trajectoryutils::LineToTube(meshIn_.getData(), glyphsize_ * glyphScale_, clipRanges,
                                           glyphOption);
    }

    meshOut_.setData(mesh);
}

}  // namespace inviwo
