/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2016-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/processors/localmaximadistance.h>
#include <modules/morphologyvisualization/util/skeletonadjacencymapper.h>
#include <modules/gromacs/processors/gromacsutil.h>

#include <glm/gtc/epsilon.hpp>
#include <inviwo/core/util/volumeramutils.h>
#include <inviwo/core/util/indexmapper.h>
#include <inviwo/core/util/assertion.h>

namespace inviwo {

using namespace skeletonutil;

//#pragma optimize("", off)

// The Class Identifier has to be globally unique. Use a reverse DNS naming scheme
const ProcessorInfo LocalMaximaDistance::processorInfo_{
    "org.inviwo.LocalMaximaDistance",  // Class identifier
    "LocalMaximaDistance",             // Display name
    "Morphology",                      // Category
    CodeState::Experimental,           // Code state
    Tags::CPU,                         // Tags
};
const ProcessorInfo LocalMaximaDistance::getProcessorInfo() const { return processorInfo_; }

LocalMaximaDistance::LocalMaximaDistance()
    : Processor()
    , gromacsInport_("gromacsIn")
    , binaryVolume_("binaryVolume")
    , distanceVolume_("distanceVolume")
    , meshIndexOutport_("outputMeshIndexCoord")
    , enableMaximaComputation_("enableLocalMaxima", "Enable Maxima Computation", false)
    , reComputeDistanceTransform_("reComputeDistanceTransform", "ReComputeDistance Transform",
                                  false)
    , normalizePositions_("normalizePositions", "Normalize Positions", true)
    , distanceScaling_("distanceScaling", "Distance Scaling", 0.1f, 0.0f, 1.0f, 0.001f)
    , maximaRadius_("maximaRadius", "Maxima Radius", 2.5f, 0.01f, 10.0f)
    , clipZ_("clipZ", "ClipZ", 0, 256, 0, 256, 1, 1)
    , colorMaxima_("colorMaxima", "Color Maxima", vec4(0.4f, 0.4f, 0.4f, 1.0f), vec4(0.0f),
                   vec4(1.0f))
    , colorMaximaInterface_("colorMaximaInterface", "Color Interface Maxima",
                            vec4(0.8f, 0.8f, 0.8f, 1.0f), vec4(0.0f), vec4(1.0f)) {
    addPort(gromacsInport_);
    addPort(binaryVolume_);
    addPort(distanceVolume_);
    addPort(meshIndexOutport_);

    colorMaxima_.setSemantics(PropertySemantics::Color);
    colorMaximaInterface_.setSemantics(PropertySemantics::Color);

    addProperty(enableMaximaComputation_);
    addProperty(reComputeDistanceTransform_);
    addProperty(normalizePositions_);
    addProperty(maximaRadius_);
    // addProperty(clipZ_);
    addProperty(colorMaxima_);
    addProperty(colorMaximaInterface_);
    // addProperty(distanceScaling_);

    binaryVolume_.onChange([&] { updateVolumeProperties(); });
    distanceVolume_.onChange([&] { updateVolumeProperties(); });

    clipZ_.onChange([&] { ski_ = nullptr; });
}

void LocalMaximaDistance::process() {
    if (!enableMaximaComputation_) return;

    if (binaryVolume_.isChanged() || distanceVolume_.isChanged()) {
        ski_ = nullptr;
    }

    bool meshDirty = false;
    if (!ski_) {
        ski_ = std::make_shared<SkeletonInfo>(binaryVolume_.getData(), distanceVolume_.getData());

        // compute general case (3D maxima) first, since interface maxima (2D) will overwrite
        // potential 3D matches
        compute3DMaxima();
        computeInterfaceMaxima();

        meshDirty = true;
    }

    if (meshDirty || colorMaxima_.isModified() || colorMaximaInterface_.isModified()) {
        generateVoxelMesh();
    }

    meshIndexOutport_.setData(meshIndexCoord_);
}

void LocalMaximaDistance::updateVolumeProperties() {
    if (binaryVolume_.getData()) {
        auto lz = 0;
        auto uz = binaryVolume_.getData()->getDimensions().z - 1;
        clipZ_.setRangeMin(lz);
        clipZ_.setRangeMax(uz);

        auto v = clipZ_.get();
        auto v1 = glm::clamp(v, ivec2(lz, lz), ivec2(uz, uz));

        if (v != v1) {
            clipZ_.setRange(v1);
        }
    }
}

void LocalMaximaDistance::compute3DMaxima() {
    auto& foregroundTest = ski_->isForeGround_;

    auto volume = distanceVolume_.getData()->getRepresentation<VolumeRAM>();
    const ivec3 dim(volume->getDimensions());

    for (int z = 1; z < dim.z - 1; ++z) {
        for (int y = 0; y < dim.y; ++y) {
            for (int x = 0; x < dim.x; ++x) {
                ivec3 centerIndex(x, y, z);
                // if (foregroundTest(centerIndex)) {
                checkForMaxima(centerIndex, volume);
                //}
            }
        }
    }
}

void LocalMaximaDistance::computeInterfaceMaxima() {
    auto& foregroundTest = ski_->isForeGround_;

    auto volume = distanceVolume_.getData()->getRepresentation<VolumeRAM>();
    const ivec3 dim(volume->getDimensions());

    auto findMaxima = [&](int zSlice) {
        for (int y = 0; y < dim.y; ++y) {
            for (int x = 0; x < dim.x; ++x) {
                ivec3 centerIndex(x, y, zSlice);
                // if (foregroundTest(centerIndex)) {
                checkFor2DMaxima(zSlice, centerIndex, volume);
                //}
            }
        }
    };

    // find 2D maxima on the first layer
    findMaxima(0);
    if (dim.z > 1) {
        // find maxima on the last layer
        findMaxima(dim.z - 1);
    }
}

void LocalMaximaDistance::checkForMaxima(const ivec3& centerIndex, const VolumeRAM* volume) {
    // check whether this voxel is a local maxima, i.e. larger than all its neighbors
    const ivec3 dim(volume->getDimensions());
    int kLower = (centerIndex.z == 0) ? 0 : -1;
    int kUpper = (centerIndex.z == dim.z - 1) ? 0 : 1;

    const auto centerValue = volume->getAsDouble(centerIndex);
    bool isMax = true;
    for (int k = kLower; k <= kUpper && isMax; ++k) {
        for (int j = -1; j <= 1 && isMax; ++j) {
            for (int i = -1; i <= 1 && isMax; ++i) {
                if ((i == 0) && (j == 0) && (k == 0)) {
                    // skip center voxel
                    continue;
                }

                ivec3 index(ivec3(i, j, k) + centerIndex);
                // ensure periodic boundary conditions
                index = (index + dim) % dim;
                const auto value = volume->getAsDouble(index);
                if (value > centerValue) {
                    // center value is no maximum, skip remaining voxel tests
                    isMax = false;
                } else if (glm::epsilonEqual(value, centerValue, glm::epsilon<double>())) {
                    if (((i < 0) || ((j < 0) && (i <= 0)) || ((k < 0) && (i <= 0) && (j <= 0)))) {
                        // there is a neighbor to the "left" which is already a maximum
                        isMax = false;
                    }
                }
            }
        }
    }
    if (isMax) {
        // local maxima found, store it in the map
        Voxel v;
        v.init();
        v.pos_ = vec3(centerIndex);
        v.index_ = ski_->index_(centerIndex);
        v.distance_ = centerValue;
        v.setVoxelFlag(VoxelFlags::ForeGround);
        v.setVoxelFlag(VoxelFlags::Maximum);
        ski_->foregroundVoxels_[v.index_] = v;
    }
}

void LocalMaximaDistance::checkFor2DMaxima(int zSlice, const ivec3& centerIndex,
                                           const VolumeRAM* volume) {
    ivwAssert(zSlice == centerIndex.x, "z slice (" << zSlice << ") does not match center index ("
                                                   << centerIndex.x
                                                   << "), (2D maxima computation)");

    // check whether this voxel is a local maxima, i.e. larger than all its neighbors
    const ivec3 dim(volume->getDimensions());

    const auto centerValue = volume->getAsDouble(centerIndex);
    bool isMax = true;
    for (int j = -1; j <= 1 && isMax; ++j) {
        for (int i = -1; i <= 1 && isMax; ++i) {
            if ((i == 0) && (j == 0)) {
                // skip center voxel
                continue;
            }

            ivec3 index(ivec3(i, j, 0) + centerIndex);
            // ensure periodic boundary conditions
            index = (index + dim) % dim;
            const auto value = volume->getAsDouble(index);
            if (value > centerValue) {
                // center value is no maximum, skip remaining voxel tests
                isMax = false;
            } else if (glm::epsilonEqual(value, centerValue, glm::epsilon<double>())) {
                if ((i < 0) || ((j < 0) && (i <= 0))) {
                    // there is a neighbor to the "left" which is already a maximum
                    isMax = false;
                }
            }
        }
    }
    if (isMax) {
        // local maxima found, store it in the map
        Voxel v;
        v.init();
        v.pos_ = vec3(centerIndex);
        v.index_ = ski_->index_(centerIndex);
        v.distance_ = centerValue;
        v.setVoxelFlag(VoxelFlags::ForeGround);
        v.setVoxelFlag(VoxelFlags::Maximum);
        v.setVoxelFlag(VoxelFlags::MaximumInterface);
        ski_->foregroundVoxels_[v.index_] = v;
    }
}

void LocalMaximaDistance::buildTree() {
    // rebuild tree
    maximaTree_.clear();
    for (auto& v : ski_->foregroundVoxels_) {
        if (v.second.hasVoxelFlag(VoxelFlags::ForeGround)) {
            if (v.second.hasVoxelFlag(VoxelFlags::Maximum)) {
                maximaTree_.insert(v.second.pos_, v.first);
            }
        }
    }
}

void LocalMaximaDistance::generateVoxelMesh() {
    if (ski_->foregroundVoxels_.empty()) {
        meshIndexCoord_ = std::make_shared<BasicMesh>();
        return;
    }

    std::shared_ptr<BasicMesh> mesh = std::make_shared<BasicMesh>();

    TransferFunction tf((int)512);
    tf.clearPoints();
    tf.addPoint(vec2(0.0f, 1.0f), vec4(1.0f, 0.0f, 0.0f, 1.0f));  // red
    tf.addPoint(vec2(0.5f, 1.0f), vec4(1.0f, 1.0f, 1.0f, 1.0f));  // white
    tf.addPoint(vec2(1.0f, 1.0f), vec4(0.0f, 0.0f, 1.0f, 1.0f));  // blue

    const vec4 colorMax3D(colorMaxima_.get());
    const vec4 colorMaxSurface(colorMaximaInterface_.get());

    VoxelMap& voxels = ski_->foregroundVoxels_;
    std::vector<BasicMesh::Vertex> vertices;
    vertices.reserve(voxels.size());
    for (auto& elem : voxels) {
        // map value from distance transform to color
        // auto color(tf.sample(elem.second.distance_));

        /*
        if (elem.second.hasVoxelFlag(VoxelFlags::Maximum)) {
            ...
        }
        */

        auto color =
            elem.second.hasVoxelFlag(VoxelFlags::MaximumInterface) ? colorMaxSurface : colorMax3D;
        // create
        vertices.push_back({ elem.second.pos_, vec3(0.0f, 1.0f, 0.0f), elem.second.pos_, color});
    }

    mesh->addVertices(vertices);

    auto ind = mesh->addIndexBuffer(DrawType::Points, ConnectivityType::None);
    std::vector<std::uint32_t> indices;
    indices.resize(voxels.size());
    std::iota(indices.begin(), indices.end(), 0);
    ind->append(indices);

    // update mesh matrices
    glm::mat4 modelMatrix, worldMatrix;
    gromacsutil::getModelAndWorldMatricesForGromacsData(*gromacsInport_.getData(), modelMatrix, worldMatrix);

    mesh->setModelMatrix(modelMatrix);
    mesh->setWorldMatrix(worldMatrix);

    meshIndexCoord_ = mesh;
}

//#pragma optimize("", on)

}  // namespace
