/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2016-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/processors/maximaedgepairs.h>
#include <modules/networkanalysis/utils/networkgraphutils.h>
#include <modules/base/algorithm/meshutils.h>

namespace inviwo {

// The Class Identifier has to be globally unique. Use a reverse DNS naming scheme
const ProcessorInfo MaximaEdgePairs::processorInfo_{
    "org.inviwo.MaximaEdgePairs",      // Class identifier
    "Maxima Edge Pairs",                // Display name
    "Morphology",              // Category
    CodeState::Experimental,  // Code state
    Tags::CPU,               // Tags
};
const ProcessorInfo MaximaEdgePairs::getProcessorInfo() const {
    return processorInfo_;
}

MaximaEdgePairs::MaximaEdgePairs()
    : Processor()
    , trajectoryMassSpringPort_("trajectoryMassSpringPort")
    , meshMaximaOut_("meshMaximaOut")
    , allEdgePairs_("allEdgePairs")
    , meshScale_("meshScale", "Mesh Scale", 1.0f, 0.01f, 10.0f)
    , valenceRadius_("valenceRadius", "Valency Radius", 1.0f, 0.01f, 10.0f)
    , plotMeshRadius_("plotMeshRadius", "Plot Mesh Radius", 1.0f, 0.01f, 10.0f)
    , showAllMaximas_("showAllMaximas", "Show All Maximas", true)
    , tf_("transferFunction", "Transfer Function")
    , logVerbosity_("logVerbosity", "Log Verbosity",
                    {{"none", "None", 0},
                     {"medium", "Medium", 1},
                     {"high", "High", 2}}) {
    addPort(trajectoryMassSpringPort_);
    addPort(meshMaximaOut_);
    addPort(allEdgePairs_);

    addProperty(meshScale_);
    addProperty(valenceRadius_);
    addProperty(plotMeshRadius_);
    addProperty(showAllMaximas_);
    addProperty(tf_);
    addProperty(logVerbosity_);
}

void MaximaEdgePairs::process() {
   
    //auto& maximaPointMesh = meshMaximaIn_.getData();

    //if (!maximaPointMesh) return;

    const vec4 colorGreen(0.2f, 0.9f, 0.2f, 1.0f);
    //const vec4 colorWhite(1.f, 1.f, 1.f, 1.f);
    //const vec4 colorBlue(0.2f, 0.2f, 0.9f, 1.0f);
    const vec4 colorRed(1.0f, 0.0f, 0.0f, 1.0f);

    //const BufferRAM* posBufferRAM = maximaPointMesh->getBuffer(0)->getRepresentation<BufferRAM>();

    //auto const& posBuffers =
    //    dynamic_cast<const Vec3BufferRAM*>(posBufferRAM)->getDataContainer();

   
    const auto& massSpringList = *trajectoryMassSpringPort_.getData();

    LabelMetaDataVectorSet massSpringListRemapped;
    trajectoryutils::massSpringToMetaDataSet(massSpringList, massSpringListRemapped);

    if (massSpringList.empty()) {
        meshMaximaOut_.setData(std::make_shared<BasicMesh>());
        return;
    }


    std::shared_ptr<BasicMesh> tmpMesh = std::make_shared<BasicMesh>();
    float radius = meshScale_.get();

    size_t vertexBufferSize = massSpringList[0]->getMaxUsedLabels() * 296;                  // 296 vertices per sphere
    tmpMesh->reserveSizeInVertexBuffer(vertexBufferSize);

    //auto massSpringListRemapped = trajectoryutils::reMapLabels(massSpringList);

    if (logVerbosity_.get() > 0) {
        std::stringstream ss;
        int mIndex = 0;
        for (auto& m : massSpringListRemapped) {
            auto sl = m; //remapped labels
            ss << "MaximaEdgePairs::Labels : " << mIndex++;
            for (auto& l : sl) {
                ss << " " << l.index_ << " ,";
            }
            ss << std::endl;
        }

        LogInfo(ss.str());
    }

    //edgepairs
    //std::map<std::pair<int, int>, std::pair<LabelMetaData, LabelMetaData>> edgePairs;
    auto edgePairsGlobal = trajectoryutils::edgePairsOfMassSpring(massSpringListRemapped);
    auto edgePairsWithWeights = std::make_shared<std::vector<vec2>>();
    for (auto& ep : edgePairsGlobal) {
        auto col = ep.second.first.color_;
        //if (ep.second.index_ >= massSpringList[0]->getMaxUsedLabels())
        //    col = ep.second.color_;
        tmpMesh = meshutil::cylinder(ep.second.first.pos_, ep.second.second.pos_, col, radius*0.25f, 16, true, tmpMesh);

        auto lp = vec2(ep.second.first.index_, ep.second.second.index_);
        edgePairsWithWeights->push_back(lp); //with duplicates
    }
    allEdgePairs_.setData(edgePairsWithWeights);
    
   

    //valence nodes    
    float maxValency = 0.0f;
    
    auto nearestMaximaLabels_remapped = trajectoryutils::valenceMaximaOfMassSpring(massSpringListRemapped, maxValency);

    //regular nodes
    for (auto& nmi : nearestMaximaLabels_remapped) {
        if (nmi.second.size()) {
            tmpMesh = meshutil::sphere(nmi.second[0].pos_, radius, nmi.second[nmi.second.size() - 1].color_, tmpMesh);
        }
    }

    for (auto& nmi : nearestMaximaLabels_remapped) {
        float valency = nmi.second.size() / maxValency;
        auto col = nmi.second[0].color_;
        col = tf_.get().sample(valency);
        tmpMesh = meshutil::sphere(nmi.second[0].pos_, valenceRadius_*valency , col, tmpMesh);

    }

    if (logVerbosity_.get() > 1) {
        std::stringstream ss;
        for (auto& v : nearestMaximaLabels_remapped) {
            ss << "Maxima : " << v.first << " - ";
            ss << "Valency : " << v.second.size() << std::endl;
        }


        int mIndex = 0;
        for (auto& m : massSpringListRemapped) {
            auto sl = m; //remapped labels
            ss << " Labels : " << mIndex++;
            for (auto& l : sl) {
                ss << " " << l.index_ << " ,";
            }
            ss << std::endl;
        }

        LogInfo(ss.str());
    }

    /*
    auto m = massSpringList[0]->getPointMesh()->getModelMatrix();
    auto w = massSpringList[0]->getPointMesh()->getWorldMatrix();
    tmpMesh->setModelMatrix(m);
    tmpMesh->setWorldMatrix(w);
    */

    networkgraphutil::setModelWorldMatrixFromMinMaxBound(tmpMesh);
    meshMaximaOut_.setData(tmpMesh);
        
    //if (logVerbosity_.get() > 1) {
    //    LogInfo("Total DepthSorted Nodes: " << (int)(ind_xy.x));
    //}

}

} // namespace

