/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2016-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/processors/maximalabelgenerator.h>
#include <modules/base/algorithm/meshutils.h>

namespace inviwo {

// The Class Identifier has to be globally unique. Use a reverse DNS naming scheme
const ProcessorInfo MaximaLabelGenerator::processorInfo_{
    "org.inviwo.MaximaLabelGenerator",      // Class identifier
    "MaximaLabelGenerator",                // Display name
    "Data Creation",              // Category
    CodeState::Experimental,  // Code state
    Tags::None,               // Tags
};
const ProcessorInfo MaximaLabelGenerator::getProcessorInfo() const {
    return processorInfo_;
}

MaximaLabelGenerator::MaximaLabelGenerator()
    : Processor()
    , meshMaximaIn_("meshMaximaIn")
    , trajectoryMassSpringPort_("trajectoryMassSpringPort")
    , meshMaximaOut_("meshMaximaOut")
    , allEdgePairs_("allEdgePairs")
    , meshScale_("meshScale", "Mesh Scale", 1.0f, 0.01f, 10.0f)
    , showAllMaximas_("showAllMaximas", "Show All Maximas", true)
    , tf_("transferFunction", "Transfer Function")
    //DOF
    , dofParameters_("dofParameters", "DOF Parameters")
    , dofProperties_("dofProperties", "DOF Options")
    , cocRadiusFilter_("cocRadiusFilter", "CoC Radius Filter", 0.1f, 0.001f, 1.0f)
    , camera_("camera", "Camera")
    , dataFrame_("dataFrame")
    , labelIndexOutPort_("labelIndexOutPort")
    , plotMesh_("plotMesh")
{
    addPort(meshMaximaIn_);
    addPort(trajectoryMassSpringPort_);
    

    addPort(meshMaximaOut_);
    addPort(allEdgePairs_);
    addPort(dataFrame_);
    addPort(plotMesh_);
    addPort(labelIndexOutPort_);

    addProperty(meshScale_);
    addProperty(showAllMaximas_);
    addProperty(tf_);

    //DOF
    addProperty(dofParameters_);
    dofParameters_.addProperty(dofProperties_);
    dofParameters_.addProperty(cocRadiusFilter_);
    addProperty(camera_);
}
    
void MaximaLabelGenerator::process() {
    updateLocalMaximaAssignment();
}


void MaximaLabelGenerator::updateLocalMaximaAssignment() {

    auto maximaPointMesh = meshMaximaIn_.getData();    

    if (!maximaPointMesh) return;

    const vec4 colorGreen(0.2f, 0.9f, 0.2f, 1.0f);
    const vec4 colorWhite(1.f, 1.f, 1.f, 1.f);
    const vec4 colorBlue(0.2f, 0.2f, 0.9f, 1.0f);
    const vec4 colorRed(1.0f, 0.0f, 0.0f, 1.0f);

    const BufferRAM* posBufferRAM = maximaPointMesh->getBuffer(0)->getRepresentation<BufferRAM>();

    auto const& posBuffers =
        dynamic_cast<const Vec3BufferRAM*>(posBufferRAM)->getDataContainer();

    std::map<size_t, size_t> nearestMaximaIndices;

    const auto& massSpringList = *trajectoryMassSpringPort_.getData();

    auto edgePairsWithWeights = std::make_shared<std::vector<vec2>>();
    
    for (auto& m : massSpringList) {
        auto sl = m->getSortedLabels();

        //compute edge pairs
        for (size_t kk = 1; kk < sl.size(); kk++) {
            auto lp = vec2(sl[kk - 1].index_, sl[kk].index_);
            edgePairsWithWeights->push_back(lp);
        }

        for (auto& l : sl) {
            if (l.index_ >= 0 && l.index_ < posBuffers.size())
                nearestMaximaIndices[l.index_]++;
        }
    }
    
    allEdgePairs_.setData(edgePairsWithWeights);

    //compute maxima valency
    std::map<int, int> nodeValencyMap;
    for (auto& nmi : nearestMaximaIndices)
        nodeValencyMap[static_cast<int>(nmi.first)] = 0;

    for (auto& ep : *edgePairsWithWeights) {
        nodeValencyMap[(int)(ep.x)] ++;
        nodeValencyMap[(int)(ep.y)] ++;
    }

    //valency count,id map
    std::vector<std::pair<int, int>> nodeValenceVectorPair;
    std::vector<int> nodeValenceVector;

    for (auto& l : nodeValencyMap) {
        nodeValenceVectorPair.push_back(std::make_pair(l.second, l.first));
        //nodeValenceVector.push_back(l.first);
    }

    //sort by valency count
    std::sort(nodeValenceVectorPair.begin(), nodeValenceVectorPair.end());

    //copy map to vector of ids
    for (auto& l : nodeValenceVectorPair)
        nodeValenceVector.push_back(l.second);
    
    //ids of 
    std::reverse(nodeValenceVector.begin(), nodeValenceVector.end());
    
    auto frameSize = 0;
    for (auto& m : massSpringList) {
        auto sl = m->getSortedLabels();
        frameSize += static_cast<int>(sl.size());
    }

    auto frame = std::make_shared<DataFrame>(frameSize);
    auto xaxis = frame->addColumn<float>("X", 0)->getTypedBuffer()->getEditableRAMRepresentation();
    auto yaxis = frame->addColumn<float>("Y", 0)->getTypedBuffer()->getEditableRAMRepresentation();
    std::array<BufferRAMPrecision<float>*, 4> color = {
        {frame->addColumn<float>("Colors R", 0)->getTypedBuffer()->getEditableRAMRepresentation(),
         frame->addColumn<float>("Colors G", 0)->getTypedBuffer()->getEditableRAMRepresentation(),
         frame->addColumn<float>("Colors B", 0)->getTypedBuffer()->getEditableRAMRepresentation(),
         frame->addColumn<float>("Colors A", 0)->getTypedBuffer()->getEditableRAMRepresentation()}};

    float xoffset = 0.0f;
    float massoffset = 200.0f;
    massoffset = nodeValenceVector.size();

    vec4 startColor(0.25f, 0.8f, 0.25f, 1.0f);
    vec4 endColor(0.8f, 0.25f, 0.25f, 1.0f);
    vec4 cocColor(0.8f, 0.2f, 0.2f, 1.0f);

    //startColor = tf_.get().sample(0.0f);
    //endColor = tf_.get().sample(1.0f);
    endColor = startColor;

    auto getValencyIdIndex = [&](int l, float& y) {
        auto nIt = std::find(nodeValenceVector.begin(), nodeValenceVector.end(), l);
        float nitpos = nIt - nodeValenceVector.begin();
        //nitpos /= nodeValenceVector.size();
        //nitpos /= nodeValenceVector.size();
        //nitpos = nodeValenceVector.size() - nitpos;
        y = nitpos;
    };


    auto plotMesh = std::make_shared<BasicMesh>();
    std::vector<std::shared_ptr<BasicMesh>> meshes;


    
    std::map<int, int> focuLabels;
   
    for (auto& m : massSpringList) {

        auto mesh = std::make_shared<BasicMesh>();
        auto inds = mesh->addIndexBuffer(DrawType::Lines, ConnectivityType::None);
        std::vector<BasicMesh::Vertex> vertices;
        auto runningIndex = 0;

        auto sortedLabels = m->getSortedLabels();
        auto xsoff = massoffset / sortedLabels.size();
        //xsoff = massoffset / (nodeValenceVector.size()+2);

        auto x = xoffset;
        vec4 col(1.0f);

        std::vector<vec3> mpoints;
        std::vector<vec3> pos;


        //dof based colors
        for (size_t lin = 0; lin < sortedLabels.size(); lin++) {
            auto l = sortedLabels[lin].index_;

            //xy coordinates
            auto y = 0.0f;
            getValencyIdIndex(l, y);

            if (l < 0 && glm::abs(l) >= posBuffers.size()) {

            }
            else if (glm::abs(l) >= posBuffers.size()) {

            }
            else {
                pos.push_back(posBuffers[l]);
            }

        }

        auto cocs = getNormalizedCOCList(pos);
        int coCIndex = 0;

        
        for (size_t lin = 0; lin < sortedLabels.size(); lin++) {
            auto l = sortedLabels[lin].index_;
            
            //xy coordinates
            auto y = 0.0f;
            getValencyIdIndex(l, y);
            xaxis->add(x);
            yaxis->add(y);

            //colors
            if (l < 0 && glm::abs(l) >= posBuffers.size()) {
                col = startColor;
            }
            else if (glm::abs(l) >= posBuffers.size()) {
                col = endColor;
            }
            else {
                col = tf_.get().sample(y / nodeValenceVector.size());
                auto cocN = cocs[coCIndex++];
                if (cocN <= cocRadiusFilter_.get()) {
                    col = cocColor;
                    focuLabels[l]++;
                }
            }

            for (size_t i = 0; i < 4; ++i) {
                color[i]->add(col[i]);
            }

            mpoints.push_back(vec3(x,y,0.0f));
            x += xsoff;
        }
        
        //populate vertices of mass spring
        uint32_t index = static_cast<uint32_t>(vertices.size());
        for (size_t mp = 1; mp < mpoints.size(); mp++) {
            inds->add(index++);
            vertices.push_back({ mpoints[mp-1], mpoints[mp-1], vec3(0.0f), vec4(vec3(0.5f),1.0f) });
            
            inds->add(index++);
            vertices.push_back({ vec3(mpoints[mp].x, mpoints[mp - 1].y, 0.0), mpoints[mp - 1], vec3(0.0f), vec4(vec3(0.5f),1.0f) });

            inds->add(index++);
            vertices.push_back({ mpoints[mp], mpoints[mp], vec3(0.0f),vec4(vec3(0.5f),1.0f) });
        }
        xoffset += massoffset;

        mesh->addVertices(vertices);
        meshes.push_back(mesh);
        
    }
    dataFrame_.setData(frame);

    //////////////////////////////////////////////////////////////////////////
    //focused lables
    auto indicesBuf_ = std::make_shared<IndexBuffer>(std::make_shared<IndexBufferRAM>());
    auto& indices = indicesBuf_->getEditableRAMRepresentation()->getDataContainer();
    
    for (auto& fl : focuLabels)
        indices.push_back(fl.first);
    labelIndexOutPort_.setData(indicesBuf_);
    //////////////////////////////////////////////////////////////////////////


    for (auto& m: meshes)
        plotMesh->append(m.get());

    plotMesh_.setData(plotMesh);



    bool maximaGeometry = false; //geometry or points
    //maximal mesh points
    if (meshMaximaIn_.isConnected() && massSpringList.size()) {
        std::shared_ptr<BasicMesh> tmpMesh = std::make_shared<BasicMesh>();
        const float radius = 0.1f * meshScale_.get();

        if (maximaGeometry) {
            size_t vertexBufferSize = posBuffers.size() * 296;                  // 296 vertices per sphere
            tmpMesh->reserveSizeInVertexBuffer(vertexBufferSize);
            

            if (showAllMaximas_) {
                for (std::size_t i = 0; i < posBuffers.size(); ++i) {
                    tmpMesh = meshutil::sphere(posBuffers[i], radius, colorGreen, tmpMesh);
                }
            }
            else {
                for (auto& nmi : nearestMaximaIndices) {
                    tmpMesh = meshutil::sphere(posBuffers[nmi.first], radius, colorGreen, tmpMesh);
                }
            }

            auto m = massSpringList[0]->getPointMesh()->getModelMatrix();
            auto w = massSpringList[0]->getPointMesh()->getWorldMatrix();
            tmpMesh->setModelMatrix(m);
            tmpMesh->setWorldMatrix(w);
        }
        else {

            auto m = massSpringList[0]->getPointMesh()->getModelMatrix();
            auto w = massSpringList[0]->getPointMesh()->getWorldMatrix();
            tmpMesh->setModelMatrix(m);
            tmpMesh->setWorldMatrix(w);

           

            size_t vertexBufferSize = posBuffers.size() * 296;                  // 296 vertices per sphere
            tmpMesh->reserveSizeInVertexBuffer(vertexBufferSize);

            std::vector<BasicMesh::Vertex> vertices;
            auto inds = tmpMesh->addIndexBuffer(DrawType::Points, ConnectivityType::None);
            inds->getDataContainer().reserve(vertexBufferSize);

            


            if (showAllMaximas_) {


                auto cocs = getNormalizedCOCList(posBuffers);

                uint32_t index = static_cast<uint32_t>(vertices.size());
                for (std::size_t i = 0; i < posBuffers.size(); ++i) {
                    auto cocN = cocs[i];
                    inds->add(index++);
                    //tmpMesh = meshutil::sphere(posBuffers[i], radius*(1.0f + mdepth), colorGreen, tmpMesh);
                    float y = cocN;
                    auto col = tf_.get().sample(y);

                    if (cocN <= cocRadiusFilter_.get())
                        col = cocColor;
                    vertices.push_back({ posBuffers[i], vec3(1.0f), vec3(0.0f), col });
                }

            }
            else {

                
                std::vector<vec4> valencyColors;
                std::vector<vec3> pos;
                for (auto& nmi : nearestMaximaIndices) {
                    pos.push_back(posBuffers[nmi.first]);
                    auto y = 0.0f;
                    getValencyIdIndex(static_cast<int>(nmi.first), y);

                    auto col = tf_.get().sample(y / nodeValenceVector.size());

                    valencyColors.push_back(col);
                }


                auto cocs = getNormalizedCOCList(pos);
                uint32_t index = static_cast<uint32_t>(vertices.size());
                for (size_t c = 0; c < pos.size(); c++) {

                    auto cocN = cocs[c];
                    auto p = pos[c];
                    auto col = valencyColors[c];
                    
                    if (cocN <= cocRadiusFilter_.get())
                        col = cocColor;

                    inds->add(index++);
                    //tmpMesh = meshutil::sphere(posBuffers[nmi.first], radius*(1.0f+mdepth), col, tmpMesh);
                    vertices.push_back({ p, vec3(1.0f), vec3(0.0f), col });
                }
            }
            tmpMesh->addVertices(vertices);
        }

        
        meshMaximaOut_.setData(tmpMesh);
    }

    /*
    {
        std::stringstream ss;
        ss << "Node valency Info ::";
        for (auto& l : nodeValenceVector)
            ss << l.first << " : " << l.second << " ,";
        LogInfo(" Labels :" << ss.str());
    }
    
    */

    /*
    {
        std::stringstream ss;        
        for (auto& m : massSpringList) {
            auto sl = m->getSortedLabels();
            for (auto& l : sl)
                ss << l << " ,";
            ss << std::endl;
        }
        LogInfo(" Labels :" << ss.str());
    }
    */


}


//for given points calculate dof parameter
std::vector<float> MaximaLabelGenerator::getNormalizedCOCList(std::vector<vec3> posBuffers) {

    const auto& massSpringList = *trajectoryMassSpringPort_.getData();
    return trajectoryutils::getNormalizedCOC(posBuffers, massSpringList, camera_, dofProperties_);
}

} // namespace

