/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2016-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/processors/maximaplots.h>
#include <modules/networkanalysis/utils/networkgraphutils.h>
#include <modules/morphologyvisualization/util/trajectoryspringmass.h>
#include <modules/base/algorithm/meshutils.h>

namespace inviwo {

// The Class Identifier has to be globally unique. Use a reverse DNS naming scheme
const ProcessorInfo MaximaPlots::processorInfo_{
    "org.inviwo.MaximaPlots",  // Class identifier
    "Maxima Plots",            // Display name
    "Morphology",              // Category
    CodeState::Experimental,   // Code state
    Tags::CPU,                 // Tags
};
const ProcessorInfo MaximaPlots::getProcessorInfo() const { return processorInfo_; }

MaximaPlots::MaximaPlots()
    : Processor()
    , trajectoryMassSpringPort_("trajectoryMassSpringPort")
    , graphInport_("graphInport")
    , jumpPlotOut_("jumpPlotOut")
    , jumpPlotSpheresOut_("jumpPlotSpheresOut")
    , jumpPlotLinesOut_("jumpPlotLinesOut")
    , circularPlot_("circularPlot")
    , circularPlotSpheresOut_("circularPlotSpheresOut")
    , circularPlotLinesOut_("circularPlotLinesOut")
    , meshScale_("meshScale", "Mesh Scale", 1.0f, 0.01f, 10.0f)
    , valenceRadius_("valenceRadius", "Valency Radius", 1.0f, 0.01f, 10.0f)
    , plotMeshRadius_("plotMeshRadius", "Plot Mesh Radius", 1.0f, 0.01f, 10.0f)
    , arcFactor_("arcFactor", "Arc Factor", 1.0f, 0.01f, 10.0f, 0.0001f)
    , circleSegments_("circleSegments", "Circle Segments", 16, 16, 512, 8)
    , arrowSegments_("circleSegments", "Circle Segments", 16, 16, 512, 8)
    , arrowSegmentProperties_("arrowSegmentProperties", "Arrow Segment Properties")
    , arrowSegmentRange_("arrowSegment", "Arrow Segment", 0.45f, 0.65f, 0.0f, 1.0f, 0.001f, 0.001f)
    , arrowThicknessScale_("arrowThicknessScale", "Arrow Thickness", 1.0f, 0.01f, 10.0f)
    , showAllMaximas_("showAllMaximas", "Show All Maximas", true)
    , showValencyDiskPlots_("showValencyDisk", "Show Valency Disk", true)
    , tf_("transferFunction", "Transfer Function")
    , verboseLogging_("logging", "Verbose Log", false) {
    // addPort(trajectoryMassSpringPort_);
    addPort(graphInport_);

    addPort(jumpPlotOut_);
    addPort(jumpPlotSpheresOut_);
    addPort(jumpPlotLinesOut_);

    addPort(circularPlot_);
    addPort(circularPlotSpheresOut_);
    addPort(circularPlotLinesOut_);

    addProperty(meshScale_);
    addProperty(valenceRadius_);
    addProperty(plotMeshRadius_);
    addProperty(arcFactor_);
    addProperty(showAllMaximas_);
    addProperty(showValencyDiskPlots_);
    addProperty(circleSegments_);

    addProperty(arrowSegmentProperties_);
    arrowSegmentProperties_.addProperty(arrowSegmentRange_);
    arrowSegmentProperties_.addProperty(arrowThicknessScale_);
    arrowSegmentProperties_.addProperty(arrowSegments_);
    addProperty(tf_);
}

void MaximaPlots::process() {
    jumpPlot();
    // circularPlot();
}

void MaximaPlots::getMapForPlotFromMassSpring(LabelMetaDataVectorSet& massSpringListRemapped) {
    bool showStatistics = verboseLogging_.get();

    const auto& massSpringList = *trajectoryMassSpringPort_.getData();

    if (!massSpringList.size()) return;

    trajectoryutils::massSpringToMetaDataSet(massSpringList, massSpringListRemapped);

    if (showStatistics) {
        float maxValency = 0.0f;
        auto nearestMaximaLabels_remapped =
            trajectoryutils::valenceMaximaOfMassSpring(massSpringListRemapped, maxValency);
        // auto& massSpringListRemapped = massSpringList;
        // auto massSpringListRemapped = trajectoryutils::reMapLabels(massSpringList);

        std::stringstream ss;
        for (auto& v : nearestMaximaLabels_remapped) {
            ss << "Maxima : " << v.first << " - ";
            ss << "Valency : " << v.second.size() << std::endl;
        }

        int mIndex = 0;
        for (auto& m : massSpringListRemapped) {
            auto sl = m;  // remapped labels
            ss << " Labels : " << mIndex++;
            for (auto& l : sl) {
                ss << " " << l.index_ << " ,";
            }
            ss << std::endl;
        }

        LogInfo(ss.str());
    }
}

void MaximaPlots::jumpPlot() {
    // auto& maximaPointMesh = meshMaximaIn_.getData();

    // if (!maximaPointMesh) return;

    const vec4 colorGreen(0.2f, 0.9f, 0.2f, 1.0f);
    // const vec4 colorWhite(1.f, 1.f, 1.f, 1.f);
    // const vec4 colorBlue(0.2f, 0.2f, 0.9f, 1.0f);
    const vec4 colorRed(1.0f, 0.0f, 0.0f, 1.0f);

    // const BufferRAM* posBufferRAM =
    // maximaPointMesh->getBuffer(0)->getRepresentation<BufferRAM>();

    // auto const& posBuffers =
    //    dynamic_cast<const Vec3BufferRAM*>(posBufferRAM)->getDataContainer();

    LabelMetaDataVectorSet massSpringListRemapped;
    // getMapForPlotFromMassSpring(massSpringListRemapped);

    NetworkGraphInfoConstSequence networkSequence = graphInport_.getVectorData();
    massSpringListRemapped = make_labelsFromGraphEdges(networkSequence);

    float maxValency = 0.0f;
    auto nearestMaximaLabels_remapped =
        trajectoryutils::valenceMaximaOfMassSpring(massSpringListRemapped, maxValency);
    // jump plot
    std::map<float, std::vector<int>> valencyLabelsDepthSortedLookup;
    for (auto& nmi : nearestMaximaLabels_remapped) {
        auto indexList = valencyLabelsDepthSortedLookup[nmi.second[0].pos_.z];

        if (std::find(indexList.begin(), indexList.end(), nmi.first) == indexList.end())
            valencyLabelsDepthSortedLookup[nmi.second[0].pos_.z].push_back(nmi.first);
    }

    // sort by valency count
    // std::sort(valencyLabels.begin(), valencyLabels.end());

    vec3 xVec(1.0f, 0.0f, 0.0f);
    vec3 yVec(0.0f, 1.0f, 0.0f);

    // auto model = massSpringList[0]->getPointMesh()->getModelMatrix();
    // auto world = massSpringList[0]->getPointMesh()->getWorldMatrix();
    // float xoffset = glm::length(model*vec4(1.0f*valencyLabelsDepthSortedLookup.size(), 0.0f,
    // 0.0f, 1.0f));  float yoffset = glm::length(model*vec4(0.0f, 1.0f*maxValency, 0.0f, 1.0f));  auto
    // model = mat4(1.0f);  auto world = mat4(1.0f);
    float xoffset =
        glm::length(vec4(1.0f * valencyLabelsDepthSortedLookup.size(), 0.0f, 0.0f, 1.0f));
    float yoffset = glm::length(vec4(0.0f, 1.0f * maxValency, 0.0f, 1.0f));

    vec2 ind_xy = vec2(0.0f);

    auto jumpPlotSpheres = std::make_shared<BasicMesh>();

    std::map<int, vec3> newMaximaPlotPoints;

    for (auto& v : valencyLabelsDepthSortedLookup) {
        auto plotIndices = v.second;
        ind_xy.y = -(plotIndices.size() / 2.0f);

        std::vector<LabelMetaData> sortedValencyIndices;

        for (auto& zDepthSortedIndex : plotIndices) {
            float valency = nearestMaximaLabels_remapped[zDepthSortedIndex].size() / maxValency;
            auto node_col = nearestMaximaLabels_remapped[zDepthSortedIndex].at(0).color_;
            if (showValencyDiskPlots_) node_col = tf_.get().sample(valency);  // valency color
            sortedValencyIndices.push_back(LabelMetaData(
                zDepthSortedIndex, vec3(valency), node_col));  // Note, this list is temporary
        }

        std::sort(sortedValencyIndices.begin(), sortedValencyIndices.end());

        for (auto& sorval : sortedValencyIndices) {
            auto zDepthSortedIndex = sorval.index_;
            // float valency = nearestMaximaLabels_remapped[zDepthSortedIndex].size() / maxValency;
            float nvalency = sorval.pos_.x;
            vec3 xyz(xVec * xoffset * ind_xy.x);
            xyz.y = (vec3(yVec * yoffset * ind_xy.y)).y;
            // jumpPlotSpheres = meshutil::sphere(xyz, (xoffset*nvalency)*plotMeshRadius_,
            // sorval.color_, jumpPlotSpheres); //valency bound - transparent

            if (showValencyDiskPlots_) {
                auto jumpPlotDisk = meshutil::disk(xyz, vec3(0.0f, 0.0f, 1.0f), sorval.color_,
                                                    (xoffset * nvalency) * plotMeshRadius_,
                                                    180);  // valency bound - transparent
                jumpPlotSpheres->append(jumpPlotDisk.get());
            }

            vec4 startEndColor = sorval.color_;
            startEndColor = nearestMaximaLabels_remapped[zDepthSortedIndex].back().color_;
            jumpPlotSpheres = meshutil::sphere(xyz, meshScale_, startEndColor,
                                                jumpPlotSpheres);  // valency bound - opaque

            newMaximaPlotPoints[zDepthSortedIndex] = xyz;
            ind_xy.y += 1.0f;
        }

        ind_xy.x += 2.0f;
    }

    if (false) {
        std::stringstream ss;
        for (auto& v : valencyLabelsDepthSortedLookup) {
            ss << "Z-Value : " << v.first << " - ";
            for (auto& ind : v.second) ss << ind << ",";

            ss << std::endl;
        }
        LogInfo(ss.str());
    }

    auto circleSegments = circleSegments_;
    auto const planeNormal = vec3(0.0f, 0.0f, 1.0f);

    auto jumpPlotSemiCircle = std::make_shared<BasicMesh>();
    auto jumpPlotSemiCircleLineTubeArrow = std::make_shared<BasicMesh>();

    for (auto& m : massSpringListRemapped) {
        auto sl = m;  // remapped labels
        for (size_t k = 1; k < sl.size(); k++) {
            if (newMaximaPlotPoints.find(sl[k - 1].index_) == newMaximaPlotPoints.end() ||
                newMaximaPlotPoints.find(sl[k].index_) == newMaximaPlotPoints.end()) {
                LogWarn("New Plot points not found");
                continue;
            }

            float valency1 = nearestMaximaLabels_remapped[sl[k - 1].index_].size() / maxValency;
            float valency2 = nearestMaximaLabels_remapped[sl[k].index_].size() / maxValency;

            auto p1 = newMaximaPlotPoints[sl[k - 1].index_];
            auto p2 = newMaximaPlotPoints[sl[k].index_];

            auto depth1 = sl[k - 1].pos_.z;
            auto depth2 = sl[k].pos_.z;

            TransferFunction tf;
            tf.clearPoints();

            if (showValencyDiskPlots_) {
                tf.addPoint(vec2(0.0f, 1.0f), tf_.get().sample(valency1));
                tf.addPoint(vec2(1.0f, 1.0f), tf_.get().sample(valency2));
            } else {
                tf.addPoint(vec2(0.0f, 1.0f), sl[k - 1].color_);
                tf.addPoint(vec2(1.0f, 1.0f), sl[k].color_);
            }

            float lineThicknessWeight = sl[k - 1].weight_;
            auto glyphOption = (trajectoryutils::LineTubeGlyphOption::AllCylinder |
                                trajectoryutils::LineTubeGlyphOption::ArrowInTheEnd);

            {
                bool switchCircleDirection = depth1 > depth2;

                if (depth1 == depth2) {
                    if (valency1 > valency2) switchCircleDirection = true;
                }

                switchCircleDirection = true;
                if (switchCircleDirection) {
                    // auto circle = trajectoryutils::semiCircle((p1 + p2) / 2.0f, p1, planeNormal,
                    // tf, circleSegments);
                    auto circle = trajectoryutils::circularArcBetweenPoints(
                        p1, p2, planeNormal, glm::length(p1 - p2) * 0.5f * arcFactor_, tf,
                        circleSegments);
                    auto& texArray = (circle->getEditableTexCoords()
                                          ->getEditableRAMRepresentation()
                                          ->getDataContainer());
                    float a = 0.0f;

                    for (auto& t : texArray) {
                        a++;
                        t = vec3(glm::mix(((float)(k - 1)) / sl.size(), ((float)(k)) / sl.size(),
                                          a / texArray.size()));
                    }

                    jumpPlotSemiCircle->append(circle.get());

                    auto segrange = arrowSegmentRange_.get();
                    auto arrowRadius = (lineThicknessWeight)*plotMeshRadius_ * arrowThicknessScale_;
                    // auto circle_low = trajectoryutils::semiCircle((p1 + p2) / 2.0f, p1,
                    // planeNormal, tf, arrowSegments_);
                    auto circle_low = trajectoryutils::circularArcBetweenPoints(
                        p1, p2, planeNormal, glm::length(p1 - p2) * 0.5f * arcFactor_, tf,
                        arrowSegments_);
                    auto tube = trajectoryutils::LineToTube(circle_low, arrowRadius, {segrange},
                                                            glyphOption);
                    jumpPlotSemiCircleLineTubeArrow->append(tube.get());
                } else {
                    // auto circle = trajectoryutils::semiCircle((p1 + p2) / 2.0f, p2, planeNormal,
                    // tf, circleSegments);
                    auto circle = trajectoryutils::circularArcBetweenPoints(
                        p1, p2, planeNormal, glm::length(p1 - p2) * 0.5f * arcFactor_, tf,
                        circleSegments);
                    auto& texArray = (circle->getEditableTexCoords()
                                          ->getEditableRAMRepresentation()
                                          ->getDataContainer());
                    float a = 0.0f;
                    for (auto& t : texArray) {
                        a++;
                        t = vec3(glm::mix(((float)(k - 1)) / sl.size(), ((float)(k)) / sl.size(),
                                          a / texArray.size()));
                    }

                    jumpPlotSemiCircle->append(circle.get());

                    auto segrange = arrowSegmentRange_.get();
                    auto arrowRadius = (lineThicknessWeight)*plotMeshRadius_ * arrowThicknessScale_;
                    // auto circle_low = trajectoryutils::semiCircle((p1 + p2) / 2.0f, p1,
                    // vec3(0.0f, 0.0f, 1.0f), tf, arrowSegments_);
                    auto circle_low = trajectoryutils::circularArcBetweenPoints(
                        p1, p2, planeNormal, glm::length(p1 - p2) * 0.5f * arcFactor_, tf,
                        arrowSegments_);
                    auto tube = trajectoryutils::LineToTube(circle_low, arrowRadius, {segrange},
                                                            glyphOption);
                    jumpPlotSemiCircleLineTubeArrow->append(tube.get());
                }
            }
        }
    }

    jumpPlotSpheres->append(jumpPlotSemiCircleLineTubeArrow.get());
    auto jumpPlotCombined = std::make_shared<BasicMesh>(*jumpPlotSpheres);
    jumpPlotCombined->append(jumpPlotSemiCircle.get());
    networkgraphutil::setModelWorldMatrixFromMinMaxBound(jumpPlotCombined);

    auto model = jumpPlotCombined->getModelMatrix();
    auto world = jumpPlotCombined->getWorldMatrix();

    {
        // combine all
        // jumpPlotSemiCircle->setModelMatrix(model);
        // jumpPlotSemiCircle->setWorldMatrix(world);
        jumpPlotOut_.setData(jumpPlotCombined);
    }

    {
        // semi circles
        jumpPlotSemiCircle->setModelMatrix(model);
        jumpPlotSemiCircle->setWorldMatrix(world);
        // networkgraphutil::setModelWorldMatrixFromMinMaxBound(jumpPlotSemiCircle);
        jumpPlotLinesOut_.setData(jumpPlotSemiCircle);
    }

    {
        // spheres
        jumpPlotSpheres->setModelMatrix(model);
        jumpPlotSpheres->setWorldMatrix(world);
        // networkgraphutil::setModelWorldMatrixFromMinMaxBound(jumpPlotSpheres);
        jumpPlotSpheresOut_.setData(jumpPlotSpheres);
    }

    // LogInfo("Total DepthSorted Nodes: " << (int)(ind_xy.x));
}

void MaximaPlots::circularPlot() {
    bool showStatistics = true;

    LabelMetaDataVectorSet massSpringListRemapped;
    getMapForPlotFromMassSpring(massSpringListRemapped);
    float maxValency = 0.0f;
    auto nearestMaximaLabels_remapped =
        trajectoryutils::valenceMaximaOfMassSpring(massSpringListRemapped, maxValency);

    // circular plot

    std::vector<vec3> maximas;
    std::vector<vec4> maximas_col;

    std::map<float, std::vector<int>> valencyLabelsDistanceSortedLookup;

    for (auto& nmi : nearestMaximaLabels_remapped) {
        auto trajectory = nmi.second;

        if (!trajectory.size()) continue;

        auto firstLabel = trajectory[0];

        auto d = glm::length(nmi.second[0].pos_ - firstLabel.pos_);

        auto xy = vec3(nmi.second[0].pos_.x, nmi.second[0].pos_.y, 0.0f);

        xy = glm::normalize(xy - firstLabel.pos_);

        d = nmi.second[0].pos_.z * glm::length(xy);

        auto indexList = valencyLabelsDistanceSortedLookup[d];

        maximas.push_back(xy + firstLabel.pos_);

        if (std::find(indexList.begin(), indexList.end(), nmi.first) == indexList.end())
            valencyLabelsDistanceSortedLookup[d].push_back(nmi.first);
    }

    vec2 minMaxDistToStartPoint = vec2(1000.0f, -1000.0f);
    for (auto& v : valencyLabelsDistanceSortedLookup) {
        minMaxDistToStartPoint.x = std::min(v.first, minMaxDistToStartPoint.x);
        minMaxDistToStartPoint.y = std::max(v.first, minMaxDistToStartPoint.y);
    }

    /*
    auto m = massSpringList[0]->getPointMesh()->getModelMatrix();
    auto w = massSpringList[0]->getPointMesh()->getWorldMatrix();
    float xoffset = glm::length(m*vec4(1.0f*valencyLabelsDistanceSortedLookup.size(), 0.0f, 0.0f,
    1.0f)); float yoffset = glm::length(m*vec4(0.0f, 1.0f*maxValency, 0.0f, 1.0f));
    */

    float xoffset =
        glm::length(vec4(1.0f * valencyLabelsDistanceSortedLookup.size(), 0.0f, 0.0f, 1.0f));
    float yoffset = glm::length(vec4(0.0f, 1.0f * maxValency, 0.0f, 1.0f));

    vec3 xVec(1.0f, 0.0f, 0.0f);
    vec3 yVec(0.0f, 1.0f, 0.0f);

    auto circularPlotRingMesh = std::make_shared<BasicMesh>();

    for (auto& v : valencyLabelsDistanceSortedLookup) {
        auto center = vec3(0.0f);
        auto startPoint = xVec * v.first;

        auto t = (v.first - minMaxDistToStartPoint.x) /
                 (minMaxDistToStartPoint.y - minMaxDistToStartPoint.x);
        auto col = tf_.get().sample(t);
        col.a = 1.0f;

        TransferFunction tf;
        tf.clearPoints();
        tf.addPoint(vec2(0.0f, 1.0f), col);
        tf.addPoint(vec2(1.0f, 1.0f), col);

        auto maximaCircle =
            trajectoryutils::circle(center, startPoint, vec3(0.0f, 0.0f, 1.0f), tf, 180);
        circularPlotRingMesh->append(maximaCircle.get());
    }

    // circularPlotRingMesh->setModelMatrix(m);
    // circularPlotRingMesh->setWorldMatrix(w);
    networkgraphutil::setModelWorldMatrixFromMinMaxBound(circularPlotRingMesh);

    vec2 ind_xy = vec2(0.0f);
    circularPlotLinesOut_.setData(circularPlotRingMesh);

    auto circularPlotTrajecMesh = std::make_shared<BasicMesh>();

    for (auto& p : maximas) {
        auto col = vec4(0.5f, 0.5f, 0.5f, 1.0f);
        circularPlotTrajecMesh = meshutil::sphere(p, meshScale_, col, circularPlotTrajecMesh);
    }

    /*
    size_t mk = 0;
    auto angle = (2.0f*M_PI)/massSpringListRemapped.size();
    auto normal = vec3(0.0f, 0.0f, 1.0f);
    auto center = vec3(0.0f);

    for (auto& m : massSpringListRemapped) {
        auto sl = m.second; //remapped labels
        for (size_t k = 0; k < sl.size(); k++) {

            float v = nearestMaximaLabels_remapped[sl[k].index_].size() / maxValency;
            auto d = glm::length(sl[k].pos_ - firstLabel.pos_);
            d = 1.0f;
            auto orth = xVec*d;
            auto startPoint = center + d * glm::normalize(glm::rotate(orth, static_cast<float>(mk *
    angle), normal)); auto col = tf_.get().sample(v);

            col = sl[k].color_;
            circularPlotTrajecMesh = meshutil::sphere(startPoint, meshScale_, col,
    circularPlotTrajecMesh);

            //{
            //circularPlotTrajecMesh = meshutil::sphere(startPoint, meshScale_, col,
    circularPlotTrajecMesh);
            //}

        }

        mk++;
    }
    */

    // trajectory spheres and tubes
    // circularPlotTrajecMesh->setModelMatrix(m);
    // circularPlotTrajecMesh->setWorldMatrix(w);
    networkgraphutil::setModelWorldMatrixFromMinMaxBound(circularPlotTrajecMesh);
    circularPlotSpheresOut_.setData(circularPlotTrajecMesh);
}

}  // namespace
