/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2017-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/processors/networkgraphs/adjacencymatrixtomesh.h>

#include <modules/gromacs/processors/gromacsutil.h>
#include <inviwo/core/datastructures/geometry/mesh.h>
#include <inviwo/core/datastructures/geometry/basicmesh.h>
#include <inviwo/core/datastructures/buffer/bufferramprecision.h>
#include <inviwo/core/datastructures/buffer/bufferram.h>
#include <inviwo/core/util/clock.h>
#include <inviwo/core/interaction/events/mouseevent.h>
#include <inviwo/core/interaction/events/pickingevent.h>
#include <modules/base/algorithm/meshutils.h>

namespace inviwo {

//#pragma optimize("", off)

// The Class Identifier has to be globally unique. Use a reverse DNS naming scheme
const ProcessorInfo AdjacencyMatrixToMesh::processorInfo_{
    "org.inviwo.AdjacencyMatrixToMesh",  // Class identifier
    "Adjacency Matrix To Mesh",          // Display name
    "Trajectory Network Graph",          // Category
    CodeState::Experimental,             // Code state
    "CPU, Morphology",                   // Tags
};
const ProcessorInfo AdjacencyMatrixToMesh::getProcessorInfo() const { return processorInfo_; }

AdjacencyMatrixToMesh::AdjacencyMatrixToMesh()
    : Processor()
    , gromacsInport_("gromacsIn")
    , maximaMeshInport_("maximaMeshIn")
    , adjacencyInport_("adjacencyIn")
    , meshOutport_("meshOut")
    , nodeProperties_("nodeprops", "Nodes", true)
    , edgeProperties_("edgeprops", "Edges", true)
    , radiusEdges_("radiusEdges", "Edge Radius", 1.0f, 0.0f, 10.0f)
    , radiusNodes_("radiusNodes", "Node Radius", 1.0f, 0.0f, 10.0f)
    , scaleNodes_("nodeScale", "Scaling", 1.0f, 0.001f, 10.0f, 0.01f)
    , scaleEdges_("edgeScale", "Scaling", 1.0f, 0.001f, 10.0f, 0.01f)
    , exponent_("exponent", "Exponent", 1.0f, 0.001f, 10.0f, 0.01f)
    , colorSource_("colorSource", "Source Color", vec4(1.0f, 0.3f, 0.3f, 1.0f), vec4(0.0f),
                   vec4(1.0f))
    , colorSink_("colorSink", "Sink Color", vec4(0.3f, 0.3f, 1.0f, 1.0f), vec4(0.0f), vec4(1.0f))
    , colorNodes_("colorNodes", "Node Color", vec4(0.6f, 0.6f, 0.6f, 1.0f), vec4(0.0f), vec4(1.0f))
    , colorEdges_("colorEdges", "Edge Color", vec4(0.8f, 0.8f, 0.8f, 1.0f), vec4(0.0f), vec4(1.0f))
    , edgeScalingMode_(
          "edgeScalingMode", "Edge Scaling",
          {{"none", "None", EdgeScaling::None}, {"jumps", "Jumps", EdgeScaling::Jumps}})
    , nodeScalingMode_("nodeScalingMode", "Node Scaling",
                       {{"none", "None", NodeScaling::None},
                        {"vertexDegreeIn", "Vertex Degree In", NodeScaling::VertexDegreeIn},
                        {"vertexDegreeOut", "Vertex Degree Out", NodeScaling::VertexDegreeOut},
                        {"vertexDegreeSum", "Vertex Degree Sum", NodeScaling::VertexDegreeSum},
                        {"vertexDegreeMax", "Vertex Degree Max", NodeScaling::VertexDegreeMax}})
    , scalingFactor_("scalingFactor", "Global Mesh Scaling", 1.0f, 0.001f, 10.0f, 0.01f)
    , enablePicking_("enablePicking", "Enable Picking", false)
    , outputMaximaID_("outputMaximaID", "Output Maxima ID", 0, 0, 190)
    , pickingMapper_(this, 1,
                     [&](PickingEvent* pobj) {
                         if (enablePicking_) pickNodeId(pobj);
                     })
    , numMaxima_(0u)
    , numEdges_(0u) {
    addPort(gromacsInport_);
    addPort(maximaMeshInport_);
    addPort(adjacencyInport_);
    addPort(meshOutport_);

    colorSource_.setSemantics(PropertySemantics::Color);
    colorSink_.setSemantics(PropertySemantics::Color);
    colorNodes_.setSemantics(PropertySemantics::Color);
    colorEdges_.setSemantics(PropertySemantics::Color);

    edgeProperties_.addProperty(colorEdges_);
    edgeProperties_.addProperty(radiusEdges_);
    edgeProperties_.addProperty(scaleEdges_);
    edgeProperties_.addProperty(exponent_);
    edgeProperties_.addProperty(edgeScalingMode_);

    nodeProperties_.addProperty(colorNodes_);
    nodeProperties_.addProperty(radiusNodes_);
    nodeProperties_.addProperty(scaleNodes_);
    nodeProperties_.addProperty(nodeScalingMode_);

    addProperty(nodeProperties_);
    addProperty(edgeProperties_);
    addProperty(enablePicking_);
    addProperty(outputMaximaID_);

    addProperty(scalingFactor_);

    addProperty(colorSource_);
    addProperty(colorSink_);

    enablePicking_.onChange([&] {
        if (enablePicking_)
            pickingMapper_.setEnabled(true);
        else
            pickingMapper_.setEnabled(false);
    });
}

void AdjacencyMatrixToMesh::process() { meshOutport_.setData(getConnectivityMesh()); }

std::shared_ptr<Mesh> AdjacencyMatrixToMesh::getConnectivityMesh() {
    auto maximaMesh = maximaMeshInport_.getData();
    // check for vec3 as position and mesh type of line or points, otherwise this processor cannot
    // handle it
    std::size_t index = 0;
    for (auto buffer : maximaMesh->getBuffers()) {
        if ((buffer.first.type == BufferType::PositionAttrib) &&
            (buffer.second->getDataFormat() == DataFormat<vec3>::get())) {
            // found a suitable position buffer
            break;
        }
        ++index;
    }
    if (index >= maximaMesh->getNumberOfBuffers()) {
        // could not find suitable buffer holding vec3 positions
        throw Exception("maxima mesh does not provide vec3 positions");
    }
    // create a copy since we want to add additional positions when splitting trajectories
    std::vector<vec3> maximaPositions =
        static_cast<const BufferRAMPrecision<vec3, BufferTarget::Data>*>(
            maximaMesh->getBuffer(index)->getRepresentation<BufferRAM>())
            ->getDataContainer();

    auto adjacency = adjacencyInport_.getData();
    auto& matrix = adjacency->matrix_;

    if (matrix.rows() != matrix.cols()) {
        throw Exception("Adjacency matrix is not square (" + std::to_string(matrix.rows()) + " x " +
                        std::to_string(matrix.cols()) + ")");
    }

    vec3 extent;
    vec3 origin;
    vec3 bBox;
    gromacsutil::getBoundingExtentForGromacsData(*gromacsInport_.getData(), extent, origin, bBox);
    const auto extentInv(vec3(1.0f) / extent);
    const auto extentHalf(extent * 0.5f);

    // create an edge for each matrix entry, i.e. m[i,j] != 0
    std::vector<int> usedMaxima(maximaPositions.size(), 0);
    std::vector<std::pair<int, int>> edgeList;
    for (int i = 0; i < matrix.rows(); ++i) {
        for (int j = 0; j < i; ++j) {
            if (matrix(i, j) || matrix(j, i)) {
                usedMaxima[i] = 1;
                usedMaxima[j] = 1;
                // apply PBC, i.e. if edge is too long split it into two parts

                const vec3 prev(maximaPositions[i]);
                const vec3 current(maximaPositions[j]);
                if (glm::any(glm::greaterThanEqual(glm::abs(current - prev), extentHalf))) {
                    // split segment
                    vec3 dir(current - prev);
                    for (int c = 0; c < 2; ++c) {
                        if (dir[c] > extentHalf[c]) {
                            dir[c] -= extent[c];
                        } else if (dir[c] < -extentHalf[c]) {
                            dir[c] += extent[c];
                        }
                    }

                    vec3 pointA(prev + dir);
                    vec3 pointB(current - dir);

                    int posIndex = static_cast<int>(maximaPositions.size());
                    maximaPositions.push_back(pointA);
                    maximaPositions.push_back(pointB);
                    edgeList.push_back({i, posIndex});
                    edgeList.push_back({posIndex + 1, j});
                } else {
                    edgeList.push_back({i, j});
                }
            }
        }
    }
    numMaxima_ = usedMaxima.size();
    numEdges_ = edgeList.size();

    return createConnectivityMesh(maximaPositions, usedMaxima, edgeList);
}

std::shared_ptr<Mesh> AdjacencyMatrixToMesh::createConnectivityMesh(
    const std::vector<vec3>& maximaPositions, const std::vector<int>& usedMaxima,
    const std::vector<std::pair<int, int>>& edgeList) {
    auto adjacency = adjacencyInport_.getData();
    auto& matrix = adjacency->matrix_;

    const size_t numMaxima =
        std::count_if(usedMaxima.begin(), usedMaxima.end(), [](auto& value) { return value != 0; });

    // create a mesh representing all used maxima and edges
    std::shared_ptr<BasicMesh> mesh = std::make_shared<BasicMesh>();

    if (!matrix.rows()) return mesh;

    const bool showNodes = nodeProperties_.isChecked();
    const bool showEdges = edgeProperties_.isChecked();

    size_t vertexBufferSize = 0;
    size_t numObjects = 0;
    if (showNodes) {
        vertexBufferSize += numMaxima * 296;  // 296 vertices per sphere
        numObjects += usedMaxima.size();
    }
    if (showEdges) {
        vertexBufferSize += edgeList.size() * 32;  // 32 vertices per cylinder
        numObjects += edgeList.size();
    }
    mesh->reserveSizeInVertexBuffer(vertexBufferSize);
    const float radiusNode = radiusNodes_.get();
    const float radiusEdge = radiusEdges_.get();

    const auto sourceColor(colorSource_.get());
    const auto sinkColor(colorSink_.get());
    const auto nodeColor(colorNodes_.get());
    const auto edgeColor(colorEdges_.get());

    pickingMapper_.resize(std::max<size_t>(1, numObjects));

    // create new pick buffer
    auto bufferRAM = std::make_shared<BufferRAMPrecision<uint32_t>>(vertexBufferSize);
    auto pickBuffer = std::make_shared<Buffer<uint32_t>>(bufferRAM);
    auto& data = bufferRAM->getDataContainer();
    auto dit = data.begin();

    auto pickingId = static_cast<uint32_t>(pickingMapper_.getPickingId(0));

    if (showNodes) {
        const auto nodeSize =
            computeNodeSizes(matrix, nodeScalingMode_.get(), radiusNodes_.get(), scaleNodes_.get());

        auto isSink = [&](int i) {
            return std::find(adjacency->sinks_.begin(), adjacency->sinks_.end(), i) !=
                   adjacency->sinks_.end();
        };
        auto isSource = [&](int i) {
            return std::find(adjacency->sources_.begin(), adjacency->sources_.end(), i) !=
                   adjacency->sources_.end();
        };

        // put all nodes into one mesh;
        for (int i = 0; i < static_cast<int>(usedMaxima.size()); ++i) {
            if (usedMaxima[i]) {
                auto color = isSource(i) ? colorSource_ : (isSink(i) ? colorSink_ : nodeColor);
                meshutil::sphere(maximaPositions[i], scalingFactor_.get() * nodeSize[i], color,
                                  mesh);

                // nodes picking ids
                for (size_t k = 0; k < 296; k++) {
                    *dit = pickingId;
                    dit++;
                }
            }
            pickingId++;
        }
    }

    // add edges
    if (showEdges) {
        const auto edgeThickness =
            computeEdgeThickness(matrix, edgeList, edgeScalingMode_.get(), radiusEdges_.get(),
                                 scaleEdges_.get(), exponent_.get());
        auto edgeIt = edgeList.begin();
        auto edgeScaleIt = edgeThickness.begin();
        for (; edgeIt != edgeList.end(); ++edgeIt, ++edgeScaleIt) {
            meshutil::cylinder(maximaPositions[edgeIt->first], maximaPositions[edgeIt->second],
                                edgeColor, scalingFactor_.get() * *edgeScaleIt, 16, false, mesh);

            // edges picking ids - assign node picking ids
            for (size_t k = 0; k < 32; k++) {
                *dit = pickingId;  // edge is assigned first nodes picking id
                dit++;
            }
            pickingId++;
        }
    }

    // replace any existing picking buffer
    mesh->addBuffer(Mesh::BufferInfo(BufferType::NumberOfBufferTypes, 4), pickBuffer);

    // update mesh matrices
    glm::mat4 modelMatrix, worldMatrix;
    gromacsutil::getModelAndWorldMatricesForGromacsData(*gromacsInport_.getData(), modelMatrix,
                                                        worldMatrix);
    mesh->setModelMatrix(modelMatrix);
    mesh->setWorldMatrix(worldMatrix);

    return mesh;
}

std::vector<float> AdjacencyMatrixToMesh::computeNodeSizes(const Eigen::MatrixXi& matrix,
                                                           NodeScaling scalingMode,
                                                           const float minSize,
                                                           const float scaling) {
    std::vector<float> nodeSize(matrix.rows(), minSize);

    switch (scalingMode) {
        case NodeScaling::None:
            break;
        case NodeScaling::VertexDegreeIn: {
            std::vector<int> colCounts(matrix.cols());
            for (int i = 0; i < matrix.cols(); ++i) {
                colCounts[i] = (matrix.col(i).array() > 0).count();
            }
            const auto maxCount =
                static_cast<float>(*std::max_element(colCounts.begin(), colCounts.end()));

            for (size_t i = 0; i < nodeSize.size(); ++i) {
                nodeSize[i] = minSize + static_cast<float>(colCounts[i]) / maxCount * scaling;
            }
            break;
        }
        case NodeScaling::VertexDegreeOut: {
            std::vector<int> rowCounts(matrix.rows());
            for (int i = 0; i < matrix.rows(); ++i) {
                rowCounts[i] = (matrix.row(i).array() > 0).count();
            }
            const auto maxCount =
                static_cast<float>(*std::max_element(rowCounts.begin(), rowCounts.end()));

            for (size_t i = 0; i < nodeSize.size(); ++i) {
                nodeSize[i] = minSize + static_cast<float>(rowCounts[i]) / maxCount * scaling;
            }
            break;
        }
        case NodeScaling::VertexDegreeSum: {
            std::vector<int> counts(matrix.rows());
            for (int i = 0; i < matrix.rows(); ++i) {
                counts[i] =
                    (matrix.row(i).array() > 0).count() + (matrix.col(i).array() > 0).count();
            }
            const auto maxCount =
                static_cast<float>(*std::max_element(counts.begin(), counts.end()));

            for (size_t i = 0; i < nodeSize.size(); ++i) {
                nodeSize[i] = minSize + static_cast<float>(counts[i]) / maxCount * scaling;
            }
            break;
        }
        case NodeScaling::VertexDegreeMax: {
            std::vector<int> counts(matrix.rows());
            for (int i = 0; i < matrix.rows(); ++i) {
                counts[i] = std::max((matrix.row(i).array() > 0).count(),
                                     (matrix.col(i).array() > 0).count());
            }
            const auto maxCount =
                static_cast<float>(*std::max_element(counts.begin(), counts.end()));

            for (size_t i = 0; i < nodeSize.size(); ++i) {
                nodeSize[i] = minSize + static_cast<float>(counts[i]) / maxCount * scaling;
            }
            break;
        }
        default:
            break;
    }
    return nodeSize;
}

std::vector<float> AdjacencyMatrixToMesh::computeEdgeThickness(
    const Eigen::MatrixXi& matrix, std::vector<std::pair<int, int>> edges, EdgeScaling scalingMode,
    const float minThickness, const float scaling, const float exponent) {
    std::vector<float> thickness(edges.size(), minThickness);
    switch (scalingMode) {
        case EdgeScaling::Jumps: {
            const auto maxCoeff = matrix.maxCoeff();

            // take care of invalid indices of split edges (> matrix.rows())
            auto it = edges.begin();
            auto dstIt = thickness.begin();
            for (; it != edges.end(); ++it, ++dstIt) {
                if (it->second >= matrix.rows()) {
                    ivwAssert(it + 1 != edges.end(), "found only part of split edge");
                    // split edge detected, recombine and fetch matrix value
                    const auto t =
                        minThickness +
                        std::pow(static_cast<float>(matrix(it->first, (it + 1)->second)) / maxCoeff,
                                 1.0f / exponent) *
                            scaling;
                    *dstIt++ = t;
                    *dstIt = t;
                    // skip second part of split edge as it was already handled
                    ++it;
                } else {
                    *dstIt = minThickness +
                             static_cast<float>(matrix(it->first, it->second)) / maxCoeff * scaling;
                }
            }
        }
        case EdgeScaling::None:
        default:
            break;
    }

    return thickness;
}

void AdjacencyMatrixToMesh::pickNodeId(PickingEvent* p) {
    const auto maximaID = p->getPickedId();

    static size_t picked = std::numeric_limits<size_t>::max();
    static bool moved = false;

    if (p->getState() == PickingState::Updated && p->getEvent()->hash() == MouseEvent::chash()) {
        auto mouseEvent = p->getEventAs<MouseEvent>();

        const bool leftMouseBtn = (mouseEvent->button() == MouseButton::Left);
        const bool mousePress = (mouseEvent->state() == MouseState::Press);
        const bool mouseRelease = (mouseEvent->state() == MouseState::Release);
        const bool mouseMove = (mouseEvent->state() == MouseState::Move);

        if (mouseMove && (mouseEvent->buttonState() & MouseButton::Left) == MouseButton::Left) {
            picked = std::numeric_limits<size_t>::max();
            moved = true;
        } else if (leftMouseBtn) {
            if (mouseRelease && (picked == maximaID) && !moved) {
                if (maximaID < numMaxima_) {
                    outputMaximaID_.set(static_cast<int>(maximaID));
                } else {
                    LogInfo("edge " << maximaID << " picked");
                }
                picked = std::numeric_limits<size_t>::max();
                // p->markAsUsed();
            } else if (mousePress) {
                moved = false;

                // p->markAsUsed();
                picked = maximaID;
            }
        }
    }
}

//#pragma optimize( "", on )

}  // namespace inviwo
