/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2017-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#ifndef IVW_ADJACENCYMATRIXTOMESH_H
#define IVW_ADJACENCYMATRIXTOMESH_H

#include <modules/morphologyvisualization/morphologyvisualizationmoduledefine.h>
#include <inviwo/core/common/inviwo.h>
#include <inviwo/core/processors/processor.h>
#include <inviwo/core/properties/optionproperty.h>
#include <inviwo/core/properties/ordinalproperty.h>
#include <inviwo/core/properties/boolcompositeproperty.h>
#include <inviwo/core/ports/imageport.h>
#include <inviwo/core/ports/meshport.h>
#include <inviwo/core/ports/datainport.h>

#include <modules/gromacs/ports/gromacport.h>
#include <modules/morphologyvisualization/datastructures/trajadjacencymatrix.h>
#include <inviwo/core/interaction/pickingmapper.h>

#include <Eigen/Eigen>

namespace inviwo {

class Mesh;

/** \docpage{org.inviwo.AdjacencyMatrixToMesh, Adjacency Matrix To Mesh}
 * ![](org.inviwo.AdjacencyMatrixToMesh.png?classIdentifier=org.inviwo.AdjacencyMatrixToMesh)
 * Convert an adjacency matrix into a mesh.
 */

/**
 * \class AdjacencyMatrixToMesh
 * \brief Convert an adjacency matrix into a mesh.
 */
class IVW_MODULE_MORPHOLOGYVISUALIZATION_API AdjacencyMatrixToMesh : public Processor {
public:
    AdjacencyMatrixToMesh();
    virtual ~AdjacencyMatrixToMesh() = default;

    virtual void process() override;

    virtual const ProcessorInfo getProcessorInfo() const override;
    static const ProcessorInfo processorInfo_;

private:
    enum class NodeScaling {
        None,
        VertexDegreeIn,
        VertexDegreeOut,
        VertexDegreeSum,
        VertexDegreeMax
    };
    enum class EdgeScaling { None, Jumps };

    std::shared_ptr<Mesh> getConnectivityMesh();
    std::shared_ptr<Mesh> createConnectivityMesh(
        const std::vector<vec3>& maximaPositions, const std::vector<int>& usedMaxima,
        const std::vector<std::pair<int, int>>& edgeList);
    static std::vector<float> computeNodeSizes(const Eigen::MatrixXi& matrix, NodeScaling scalingMode,
                                               const float minSize, const float scaling);
    static std::vector<float> computeEdgeThickness(const Eigen::MatrixXi& matrix,
                                                   std::vector<std::pair<int, int>> edges,
                                                   EdgeScaling scalingMode, const float minThickness,
                                                   const float scaling, const float exponent);
    void pickNodeId(PickingEvent *p);

    GromacsMultiInportFlat gromacsInport_;
    MeshInport maximaMeshInport_;
    DataInport<TrajAdjacencyMatrix> adjacencyInport_;
    MeshOutport meshOutport_;

    BoolCompositeProperty nodeProperties_;
    BoolCompositeProperty edgeProperties_;

    FloatProperty radiusNodes_;
    FloatProperty radiusEdges_;
    FloatProperty scaleNodes_;
    FloatProperty scaleEdges_;
    FloatProperty exponent_;

    FloatVec4Property colorSource_;
    FloatVec4Property colorSink_;
    FloatVec4Property colorNodes_;
    FloatVec4Property colorEdges_;

    TemplateOptionProperty<EdgeScaling> edgeScalingMode_;
    TemplateOptionProperty<NodeScaling> nodeScalingMode_;
    FloatProperty scalingFactor_;
    IntProperty outputMaximaID_;
    BoolProperty enablePicking_;

    //Picking
    PickingMapper pickingMapper_;
    // used for translating picking IDs back to maxima IDs and edges
    size_t numMaxima_;
    size_t numEdges_;
};

}  // namespace inviwo

#endif  // IVW_ADJACENCYMATRIXTOMESH_H
