﻿/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2017-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/processors/networkgraphs/chargepoolvis.h>

#include <modules/gromacs/processors/gromacsutil.h>

#include <inviwo/core/datastructures/geometry/mesh.h>
#include <inviwo/core/datastructures/geometry/basicmesh.h>
#include <inviwo/core/datastructures/buffer/bufferramprecision.h>
#include <inviwo/core/datastructures/buffer/bufferram.h>
#include <inviwo/core/util/clock.h>
#include <inviwo/core/interaction/events/keyboardevent.h>
#include <inviwo/core/interaction/events/eventmatcher.h>
#include <modules/eigenutils/eigenutils.h>
#include <modules/base/algorithm/meshutils.h>

#include <Eigen/Core>
#include <Eigen/Eigen>
#include <Eigen/SVD>

#include <random>

namespace inviwo {

namespace util {

// this is a temporary fix for Eigen matrix to glm mat conversion
template <
    typename T, unsigned Rows, unsigned Cols,
    typename std::enable_if<(Rows >= 2 && Rows <= 4 && Cols >= 2 && Cols <= 4), int>::type = 0>
auto eigen2glmFixed(const Eigen::Matrix<T, Rows, Cols>& m) {
    using GlmMatrix = typename util::glmtype<T, Cols, Rows>::type;
    GlmMatrix outm;
    for (size_t i = 0; i < Cols; i++) {
        for (size_t j = 0; j < Rows; j++) {
            outm[i][j] = m(i, j);
        }
    }
    return outm;
}

}  // namespace util

//#pragma optimize("", off)

// The Class Identifier has to be globally unique. Use a reverse DNS naming scheme
const ProcessorInfo ChargePoolVis::processorInfo_{
    "org.inviwo.ChargePoolVis",  // Class identifier
    "Charge Pool Vis",           // Display name
    "Trajectory Network Graph",  // Category
    CodeState::Experimental,     // Code state
    Tags::CPU,                   // Tags
};
const ProcessorInfo ChargePoolVis::getProcessorInfo() const { return processorInfo_; }

ChargePoolVis::ChargePoolVis()
    : Processor()
    , gromacsInport_("gromacsIn")
    , maximaVolInport_("maximaVolumeIn")
    , maximaMeshInport_("maximaMeshIn")
    , adjacencyInport_("adjacencyIn")
    , trajectorySmsInport_("trajectorySpringMassIn")
    , meshOutport_("meshOut")
    , maximaPointsOutport_("maximaPointsOut")
    , nodeProperties_("nodeprops", "Nodes", true)
    , radiusNodes_("radiusNodes", "Node Radius", 1.0f, 0.0f, 10.0f)
    , scaleNodes_("nodeScale", "Scaling", 1.0f, 0.001f, 10.0f, 0.01f)
    , colorSource_("colorSource", "Source Color", vec4(1.0f, 0.3f, 0.3f, 1.0f), vec4(0.0f),
                   vec4(1.0f))
    , colorSink_("colorSink", "Sink Color", vec4(0.3f, 0.3f, 1.0f, 1.0f), vec4(0.0f), vec4(1.0f))
    , colorNodes_("colorNodes", "Node Color", vec4(0.6f, 0.6f, 0.6f, 1.0f), vec4(0.0f), vec4(1.0f))
    , colorSelection_("colorSelection", "Selection Color", vec4(1.0f, 0.85f, 0.25f, 1.0f),
                      vec4(0.0f), vec4(1.0f))
    , colorPos_("colorPos", "Position Color", vec4(1.0f, 0.85f, 0.1f, 1.0f), vec4(0.0f), vec4(1.0f))
    , nodeScalingMode_(
          "nodeScalingMode", "Node Scaling",
          {{"none", "None", NodeScaling::None},
           {"vertexDegreeIn", "Vertex Degree In", NodeScaling::VertexDegreeIn},
           {"vertexDegreeOut", "Vertex Degree Out", NodeScaling::VertexDegreeOut},
           {"vertexDegreeSum", "Vertex Degree Sum", NodeScaling::VertexDegreeSum},
           {"vertexDegreeMax", "Vertex Degree Max", NodeScaling::VertexDegreeMax},
           {"maximaSize", "Maxima Size", NodeScaling::MaximaSize},
           {"numberOfCharges", "Number of Charges", NodeScaling::NumberOfCharges},
           {"normalizedCharges", "Normalized Charges", NodeScaling::NormalizedCharges}})
    , meshScalingFactor_("scalingFactor", "Global Mesh Scaling", 1.0f, 0.001f, 10.0f, 0.01f)

    , ellipsoidProperties_("ellipsoidprops", "Ellipsoids")
    , pcaDomain_("pcaDomain", "PCA Domain",
                 {{"morphology", "Morphology", PCADomain::Morphology},
                  {"charges", "Charges", PCADomain::Charges}})
    , minimumThickness_("minimumThickness", "Min Thickness", 0.1f, 0.001f, 1.0f, 0.01f)
    , minChargeCount_("minChargeCount", "Min Charge Count per Maxima", 3, 3, 50)
    , checkValency_("checkValency", "Check Valency", false)
    , minValency_("minValency", "Min Valency", 0, 0, 20)
    , pcaNormalization_("pcaNormalization", "PCA Normalization",
                        {{"none", "None", Normalization::None},
                         {"percentage", "Percentage", Normalization::Percentage},
                         {"eigenvalue", "Largest Eigenvalue", Normalization::LargestEigenvalue},
                         {"volume", "Volume", Normalization::Volume}},
                        2)
    , highlightSelectedMaxima_("highlightSelectedMaxima", "Highlight Selected Maxima", true)
    , outputMaximaID_("outputMaximaID", "Output Maxima ID", 0, 0, 190)
    , jigglePos_("jigglePos", "Jiggle Positions", true)
    , keyAdvanceSelectedID_("keyAdvanceID", "Advance ID",
                            [this](Event* e) {
                                outputMaximaState_.dirty = true;
                                outputMaximaState_.newValue = outputMaximaID_.get() + 1;
                                this->invalidate(InvalidationLevel::InvalidOutput);
                            },
                            std::make_unique<GeneralEventMatcher>([](Event* e) {
                                if (e->hash() == KeyboardEvent::chash()) {
                                    auto keyEvent = static_cast<KeyboardEvent*>(e);
                                    if (keyEvent->state() == KeyState::Release) return false;
                                    return keyEvent->key() == IvwKey::Plus ||
                                           keyEvent->key() == IvwKey::Right;
                                }
                                return false;
                            }))
    , keyDecreaseSelectedID_("keyDecreaseID", "Decrease ID",
                             [this](Event* e) {
                                 if (outputMaximaID_.get() > 0) {
                                     outputMaximaState_.dirty = true;
                                     outputMaximaState_.newValue = outputMaximaID_.get() - 1;
                                     this->invalidate(InvalidationLevel::InvalidOutput);
                                 }
                             },
                             std::make_unique<GeneralEventMatcher>([](Event* e) {
                                 if (e->hash() == KeyboardEvent::chash()) {
                                     auto keyEvent = static_cast<KeyboardEvent*>(e);
                                     if (keyEvent->state() == KeyState::Release) return false;
                                     return keyEvent->key() == IvwKey::Minus ||
                                            keyEvent->key() == IvwKey::Left;
                                 }
                                 return false;
                             })) {
    addPort(gromacsInport_);
    addPort(maximaVolInport_);
    addPort(maximaMeshInport_);
    addPort(adjacencyInport_);
    addPort(trajectorySmsInport_);
    addPort(meshOutport_);
    addPort(maximaPointsOutport_);

    colorSource_.setSemantics(PropertySemantics::Color);
    colorSink_.setSemantics(PropertySemantics::Color);
    colorNodes_.setSemantics(PropertySemantics::Color);
    colorSelection_.setSemantics(PropertySemantics::Color);
    colorPos_.setSemantics(PropertySemantics::Color);

    nodeProperties_.addProperty(colorNodes_);
    nodeProperties_.addProperty(radiusNodes_);
    nodeProperties_.addProperty(scaleNodes_);
    nodeProperties_.addProperty(nodeScalingMode_);

    addProperty(nodeProperties_);

    ellipsoidProperties_.addProperty(pcaDomain_);
    ellipsoidProperties_.addProperty(minimumThickness_);
    ellipsoidProperties_.addProperty(minChargeCount_);
    ellipsoidProperties_.addProperty(checkValency_);
    ellipsoidProperties_.addProperty(minValency_);
    ellipsoidProperties_.addProperty(pcaNormalization_);

    addProperty(ellipsoidProperties_);

    addProperty(meshScalingFactor_);
    addProperty(colorSource_);
    addProperty(colorSink_);

    addProperty(highlightSelectedMaxima_);
    addProperty(outputMaximaID_);
    addProperty(jigglePos_);
    addProperty(colorSelection_);
    addProperty(colorPos_);

    keyAdvanceSelectedID_.setVisible(false);
    keyDecreaseSelectedID_.setVisible(false);
    addProperty(keyAdvanceSelectedID_);
    addProperty(keyDecreaseSelectedID_);
}

void ChargePoolVis::process() {
    if (maximaVolInport_.getData()->getDataFormat() != DataFormat<glm::int16>::get()) {
        throw Exception("ChargePoolVis processor only supports int16 maxima volumes");
    }

    if (outputMaximaState_.dirty) {
        outputMaximaID_.set(outputMaximaState_.newValue);
        outputMaximaState_.dirty = false;
    }

    auto meshes = getConnectivityMesh();
    meshOutport_.setData(meshes.first);
    maximaPointsOutport_.setData(meshes.second);
}

std::pair<std::shared_ptr<Mesh>, std::shared_ptr<Mesh>> ChargePoolVis::getConnectivityMesh() const {

    mat4 d;
    d[0][4] = 4.0f;


    const double minEigenValue = minimumThickness_.get();
    // minimum number of charges of a maxima to draw an ellipsoid instead of a sphere
    const int minCharges = std::max(3, minChargeCount_.get());

    auto volume = static_cast<const VolumeRAMPrecision<glm::int16>*>(
        maximaVolInport_.getData()->getRepresentation<VolumeRAM>());
    auto volData = volume->getDataTyped();

    auto maximaMesh = maximaMeshInport_.getData();
    // check for vec3 as position and mesh type of line or points, otherwise this processor cannot
    // handle it
    std::size_t index = 0;
    for (auto buffer : maximaMesh->getBuffers()) {
        if ((buffer.first.type == BufferType::PositionAttrib) &&
            (buffer.second->getDataFormat() == DataFormat<vec3>::get())) {
            // found a suitable position buffer
            break;
        }
        ++index;
    }
    if (index >= maximaMesh->getNumberOfBuffers()) {
        // could not find suitable buffer holding vec3 positions
        throw Exception("maxima mesh does not provide vec3 positions");
    }
    // create a copy since we want to add additional positions when splitting trajectories
    const auto& maximaPositions = static_cast<const BufferRAMPrecision<vec3, BufferTarget::Data>*>(
                                      maximaMesh->getBuffer(index)->getRepresentation<BufferRAM>())
                                      ->getDataContainer();

    const auto maximaCount = maximaPositions.size();

    const ivec3 dim(volume->getDimensions());
    util::IndexMapper<3, int> im(dim);

    const dvec3 ddim(dim);
    const dvec3 ddimInv(1.0 / ddim);

    std::vector<int> maxVolSizes(maximaCount);  // size of each maxima in voxels
    {
        IVW_CPU_PROFILING_IF(150, "compute maxima volume sizes");
        for (int i = 0; i < dim.x * dim.y * dim.z; ++i) {
            if (volData[i] > 0) {
                ++maxVolSizes[volData[i] - 1];
            }
        }

        for (auto elem : maxVolSizes) {
            if (elem == 0) {
                throw Exception("empty maxima region");
            }
        }
    }

    vec3 extent;
    vec3 origin;
    vec3 bBox;
    gromacsutil::getBoundingExtentForGromacsData(*gromacsInport_.getData(), extent, origin, bBox);
    const auto extentInv(vec3(1.0f) / extent);
    const auto extentHalf(extent * 0.5f);

    std::vector<std::vector<dvec3>> chargePool(maximaCount);  // charge positions per maxima
    std::vector<int> chargePoolCounts;
    if (pcaDomain_.get() == PCADomain::Morphology) {
        IVW_CPU_PROFILING_IF(150, "computing morphology pools");

        // assign all voxels of the volume to their respective morphology pools
        for (int i = 0; i < dim.x * dim.y * dim.z; ++i) {
            auto maximaID = volData[i];
            if (maximaID > 0) {
                // add voxel position to matching pool
                auto p = vec3(im(i));
                // unapply PBC
                vec3 dir(vec3(p) - maximaPositions[maximaID - 1]);
                if (glm::any(glm::greaterThanEqual(glm::abs(dir), extentHalf))) {
                    for (int c = 0; c < 2; ++c) {
                        if (dir[c] > extentHalf[c]) {
                            p[c] -= extent[c];
                        } else if (dir[c] < -extentHalf[c]) {
                            p[c] += extent[c];
                        }
                    }
                }
                chargePool[maximaID - 1].push_back(p);
            }
        }
        // copy voxel counts to chargePool counts
        chargePoolCounts.assign(maxVolSizes.begin(), maxVolSizes.end());
    } else {
        IVW_CPU_PROFILING_IF(150, "building charge pools");

        auto applyPBC = [&](const auto& p) {
            return ddim * glm::fract(p * ddimInv);
        };

        auto fetchMaxima = [&](const auto& p) {
            return volData[im(ivec3(p + 0.5))];
        };

        std::size_t errorCount = 0u;
        auto springMassSystems = trajectorySmsInport_.getData();
        if (!springMassSystems->empty()) {
            for (auto sms : *springMassSystems) {
                const auto numPos = sms->getPositions().size();

                auto it = sms->getPositions().begin();
                for (; it != sms->getPositions().end(); ++it) {
                    // apply periodic boundary conditions
                    auto p = applyPBC(*it);
                    auto maximaID = fetchMaxima(p);
                    if (maximaID != 0) {
                        // unapply PBC
                        vec3 dir(vec3(p) - maximaPositions[maximaID - 1]);
                        if (glm::any(glm::greaterThanEqual(glm::abs(dir), extentHalf))) {
                            for (int c = 0; c < 2; ++c) {
                                if (dir[c] > extentHalf[c]) {
                                    p[c] -= extent[c];
                                } else if (dir[c] < -extentHalf[c]) {
                                    p[c] += extent[c];
                                }
                            }
                        }

                        // reduce maxima indices by 1 as they are off by one,
                        // save position in matching charge pool
                        chargePool[maximaID - 1].push_back(p);
                    } else {
                        // zero indicates no maxima found, something is wrong
                        ++errorCount;
                    }
                }
            }
        }
        if (errorCount > 0) {
            LogWarn("could not locate maxima for " << errorCount << " point(s)");
        }

        // accumulate all charges per pool
        std::transform(chargePool.begin(), chargePool.end(), std::back_inserter(chargePoolCounts),
                       [](auto& vec) { return static_cast<int>(vec.size()); });
    }
    /*
    std::ofstream of("h:/Documents/projects/morphology/mathematica/chargepool.txt");
    auto printVector = [&](const std::string& name, const auto& data) {
        of << name << " = {\n";
        std::copy(data.begin(), data.end(), util::make_ostream_joiner(of, ", "));
        of << "};\n";
    };
    printVector("poolCounts", chargePoolCounts);
    printVector("poolSize", maxVolSizes);
    of.close();
    */
    auto adjacency = adjacencyInport_.getData();
    auto& matrix = adjacency->matrix_;

    if (matrix.rows() != matrix.cols()) {
        throw Exception("Adjacency matrix is not square (" + std::to_string(matrix.rows()) + " x " +
                        std::to_string(matrix.cols()) + ")");
    }

    std::vector<mat3> basisVectors;  // per maxima
    {
        IVW_CPU_PROFILING_IF(150, "perform PCA");

        for (size_t i = 0; i < chargePool.size(); ++i) {
            auto& data = chargePool[i];
            if (data.size() < minCharges) {
                basisVectors.push_back(mat3(minEigenValue));
                continue;
            }

            if (checkValency_.get()) {
                // check the valency/degree of the current maximum, i.e. the number of incident
                // edges
                auto valency = (matrix.col(i).array() > 0).count();
                if (valency < minValency_.get()) {
                    basisVectors.push_back(mat3(minEigenValue));
                    continue;
                }
            }

            // convert vector of vec3 into a Eigen matrix
            Eigen::MatrixXd m(data.size(), 3);
            for (int j = 0; j < static_cast<int>(data.size()); ++j) {
                m.row(j) = Eigen::VectorXd::Map(glm::value_ptr(data[j]), 3);
            }

            // compute eigenvectors using singular value decomposition (SVD)
            //
            // SVD only equivalent to PCA of covariance if matrix is centered, i.e. column means are
            // zero matrix A which is n x p with n samples of p variables (per row)
            //
            // solves A = U S V*  where S is a diagonal matrix (singular values), V holds the right
            // singular vectors (p x p)
            //
            // eigenvalues correspond to λ_i = s^2_i / (n−1)
            //
            // see
            // http://stats.stackexchange.com/questions/134282/relationship-between-svd-and-pca-how-to-use-svd-to-perform-pca
            //

            // center matrix
            Eigen::Vector3d mean = m.colwise().mean();
            m = m.rowwise() - mean.transpose();

            dvec3 offset(util::eigen2glm(mean));

            Eigen::JacobiSVD<Eigen::MatrixXd> svd(m, Eigen::ComputeThinU | Eigen::ComputeThinV);

            if ((svd.matrixV().cols() != 3) || (svd.matrixV().rows() != 3)) {
                throw Exception("cannot compute eigenvectors, too few samples");
            }

            dmat3 basis(util::eigen2glm<double, 3, 3>(svd.matrixV()));

            // scale basis according to eigenvalues
            Eigen::Vector3d ev = (svd.singularValues().array().square() / (static_cast<double>(m.rows()) - 1.0)).sqrt();
            dvec3 eigenvalues =
                util::eigen2glm<double, 3, 1>(ev);

            switch (pcaNormalization_.get()) {
                case Normalization::Percentage:
                    eigenvalues /= glm::compAdd(eigenvalues);
                    break;
                case Normalization::LargestEigenvalue:
                    // normalize with respect to largest eigenvalue
                    eigenvalues /= eigenvalues.x;
                    break;
                case Normalization::Volume:
                    // normalize with respect to maxima volume
                    eigenvalues /= static_cast<double>(maxVolSizes[i]);
                    break;
                case Normalization::None:
                default:
                    break;
            }

            for (int j = 0; j < 3; ++j) {
                basis[j] = basis[j] * std::max(eigenvalues[j], minEigenValue);
            }

            basisVectors.push_back(mat3(basis));
        }
    }

    // create an edge for each matrix entry, i.e. m[i,j] != 0
    std::vector<int> usedMaxima(maximaPositions.size(), 0);
    for (int i = 0; i < matrix.rows(); ++i) {
        for (int j = 0; j < i; ++j) {
            if (matrix(i, j) || matrix(j, i)) {
                usedMaxima[i] = 1;
                usedMaxima[j] = 1;
            }
        }
    }

    auto pointMesh = createMaximaPointMesh(chargePool, chargePoolCounts, outputMaximaID_.get());
    auto connectivityMesh = createConnectivityMesh(maximaPositions, usedMaxima, maxVolSizes,
                                                   chargePoolCounts, basisVectors);
    return {connectivityMesh, pointMesh};
}

std::shared_ptr<Mesh> ChargePoolVis::createConnectivityMesh(
    const std::vector<vec3>& maximaPositions, const std::vector<int>& usedMaxima,
    const std::vector<int>& poolVoxelSizes, const std::vector<int>& chargePoolCounts,
    const std::vector<mat3>& basisVectors) const {
    auto adjacency = adjacencyInport_.getData();
    auto& matrix = adjacency->matrix_;

    const size_t numMaxima =
        std::count_if(usedMaxima.begin(), usedMaxima.end(), [](auto& value) { return value != 0; });

    // create a mesh representing all used maxima
    std::shared_ptr<BasicMesh> mesh = std::make_shared<BasicMesh>();

    size_t vertexBufferSize = numMaxima * 296;  // 296 vertices per sphere

    mesh->reserveSizeInVertexBuffer(vertexBufferSize);
    const float radiusNode = radiusNodes_.get();

    const auto sourceColor(colorSource_.get());
    const auto sinkColor(colorSink_.get());
    const auto nodeColor(colorNodes_.get());

    const auto nodeSize =
        computeNodeSizes(matrix, poolVoxelSizes, chargePoolCounts, nodeScalingMode_.get(),
                         radiusNodes_.get(), scaleNodes_.get());

    auto isSink = [&](int i) {
        return std::find(adjacency->sinks_.begin(), adjacency->sinks_.end(), i) !=
               adjacency->sinks_.end();
    };
    auto isSource = [&](int i) {
        return std::find(adjacency->sources_.begin(), adjacency->sources_.end(), i) !=
               adjacency->sources_.end();
    };

    auto deformedSphere = [&](auto& basis, auto& offset, float scaling, const vec4& color) {
#if 1
        auto sphere = meshutil::sphere(vec3(0.0f), 1.0f, color);

        auto& vertices =
            sphere->getEditableVertices()->getEditableRAMRepresentation()->getDataContainer();
        auto& normals =
            sphere->getEditableNormals()->getEditableRAMRepresentation()->getDataContainer();

        mat4 m(basis * scaling);
        m[3] = vec4(offset, 1.0f);

        mat3 normalMatrix(glm::inverseTranspose(mat3(m)));

        for (auto& vertex : vertices) {
            vertex = vec3(m * vec4(vertex, 1.0f));
        }
        for (auto& normal : normals) {
            normal = normalMatrix * normal;
        }
        return sphere;
#else

        float r = scaling * 0.03f;
        auto mesh = meshutil::arrow(offset, offset + basis[0] * scaling, vec4(1.0f, 0.0f, 0.0f, 1.0f), r, 0.15f, r*1.5f);
        mesh->append(meshutil::arrow(offset, offset + basis[1] * scaling, vec4(0.0f, 1.0f, 0.0f, 1.0f), r, 0.15f, r*1.5f).get());
        mesh->append(meshutil::arrow(offset, offset + basis[2] * scaling, vec4(0.0f, 0.0f, 1.0f, 1.0f), r, 0.15f, r*1.5f).get());
        return mesh;
#endif
    };

    // put all nodes into one mesh;
    for (int i = 0; i < static_cast<int>(usedMaxima.size()); ++i) {
        if (usedMaxima[i]) {
            auto color = isSource(i) ? colorSource_ : (isSink(i) ? colorSink_ : nodeColor);
            if (highlightSelectedMaxima_.get() && i == outputMaximaID_.get()) {
                color = colorSelection_.get();
            }

            auto sphere = deformedSphere(basisVectors[i], maximaPositions[i],
                                         meshScalingFactor_.get() * nodeSize[i], color);
            mesh->append(sphere.get());
        }
    }

    // update mesh matrices
    glm::mat4 modelMatrix, worldMatrix;
    gromacsutil::getModelAndWorldMatricesForGromacsData(*gromacsInport_.getData(), modelMatrix,
                                                        worldMatrix);
    mesh->setModelMatrix(modelMatrix);
    mesh->setWorldMatrix(worldMatrix);

    return mesh;
}

std::shared_ptr<Mesh> ChargePoolVis::createMaximaPointMesh(
    const std::vector<std::vector<dvec3>>& chargePool, const std::vector<int>& counts,
    int pickedID) const {

    if (!highlightSelectedMaxima_.get() || (pickedID < 0) || (pickedID >= chargePool.size())) {
        // invalid input index
        return std::make_shared<BasicMesh>();
    }
    auto mesh = std::make_shared<Mesh>(DrawType::Points, ConnectivityType::None);

    std::vector<vec3> positions(chargePool[pickedID].size());
    std::vector<vec4> colors(chargePool[pickedID].size(), colorPos_.get());

    // apply jiggling to all positions
    std::mt19937 rand;
    std::uniform_real_distribution<float> dis(-0.5f, 0.5f);
    rand.seed(static_cast<std::mt19937::result_type>(0));

    auto randVec3 = [&]() { return vec3(dis(rand), dis(rand), dis(rand)); };

    if (jigglePos_.get()) {
        std::transform(chargePool[pickedID].begin(), chargePool[pickedID].end(), positions.begin(),
                       [&](const dvec3& p) -> vec3 { return vec3(p) + randVec3(); });
    }
    else {
        std::transform(chargePool[pickedID].begin(), chargePool[pickedID].end(), positions.begin(),
                       [](const dvec3& p) -> vec3 { return vec3(p); });
    }

    mesh->addBuffer(BufferType::PositionAttrib, util::makeBuffer(std::move(positions)));
    mesh->addBuffer(BufferType::ColorAttrib, util::makeBuffer(std::move(colors)));

    // update mesh matrices
    glm::mat4 modelMatrix, worldMatrix;
    gromacsutil::getModelAndWorldMatricesForGromacsData(*gromacsInport_.getData(), modelMatrix,
                                                        worldMatrix);
    mesh->setModelMatrix(modelMatrix);
    mesh->setWorldMatrix(worldMatrix);

    return mesh;
}

std::vector<float> ChargePoolVis::computeNodeSizes(const Eigen::MatrixXi& matrix,
                                                   const std::vector<int>& poolVoxelSizes,
                                                   const std::vector<int>& chargePoolCounts,
                                                   NodeScaling scalingMode, const float minSize,
                                                   const float scaling) {
    std::vector<float> nodeSize(matrix.cols(), minSize);

    if (scalingMode == NodeScaling::None) {
        return nodeSize;
    }

    std::vector<float> values(matrix.cols());

    switch (scalingMode) {
        case NodeScaling::None:
            break;
        case NodeScaling::VertexDegreeIn: {
            // count column entries
            for (int i = 0; i < matrix.cols(); ++i) {
                values[i] = (matrix.col(i).array() > 0).count();
            }
            break;
        }
        case NodeScaling::VertexDegreeOut: {
            // count row entries
            for (int i = 0; i < matrix.rows(); ++i) {
                values[i] = (matrix.row(i).array() > 0).count();
            }
            break;
        }
        case NodeScaling::VertexDegreeSum: {
            // build sum of row and column count entries for each maxima
            for (int i = 0; i < matrix.rows(); ++i) {
                values[i] =
                    (matrix.row(i).array() > 0).count() + (matrix.col(i).array() > 0).count();
            }
            break;
        }
        case NodeScaling::VertexDegreeMax: {
            // build max of row and column count entries for each maxima
            for (int i = 0; i < matrix.rows(); ++i) {
                values[i] = std::max((matrix.row(i).array() > 0).count(),
                                     (matrix.col(i).array() > 0).count());
            }
            break;
        }
        case NodeScaling::MaximaSize:
            std::transform(poolVoxelSizes.begin(), poolVoxelSizes.end(), values.begin(),
                           [](int value) { return static_cast<float>(value); });
            break;
        case NodeScaling::NumberOfCharges:
            std::transform(chargePoolCounts.begin(), chargePoolCounts.end(), values.begin(),
                           [](int value) { return static_cast<float>(value); });
            break;
        case NodeScaling::NormalizedCharges:
            std::transform(chargePoolCounts.begin(), chargePoolCounts.end(), poolVoxelSizes.begin(),
                           values.begin(), [](auto count, auto size) {
                               return static_cast<float>(count) / static_cast<float>(size);
                           });
            {
                const auto maxValue =
                    static_cast<float>(*std::max_element(values.begin(), values.end()));
                LogInfoCustom("chargepoolvis", "max value normalized charges: " << maxValue);
            }
            break;
        default:
            break;
    }

    const auto maxValue = static_cast<float>(*std::max_element(values.begin(), values.end()));
    const float factor = scaling / maxValue;
    for (size_t i = 0; i < nodeSize.size(); ++i) {
        nodeSize[i] = minSize + static_cast<float>(values[i]) * factor;
    }

    return nodeSize;
}

//#pragma optimize("", on)

}  // namespace inviwo
