/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2017-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/processors/networkgraphs/graphtocircularplotmesh.h>
#include <modules/morphologyvisualization/util/trajectorylabelmetadata.h>
#include <modules/networkanalysis/utils/networkgraphutils.h>
#include <modules/morphologyvisualization/util/trajectoryspringmass.h>

#pragma optimize("", off)

using namespace ogdf;

namespace inviwo {

// The Class Identifier has to be globally unique. Use a reverse DNS naming scheme
const ProcessorInfo GraphToCircularPlotMesh::processorInfo_{
    "org.inviwo.GraphToCircularPlotMesh",  // Class identifier
    "Graph To Circular Plot Mesh",         // Display name
    "Morphology",                          // Category
    CodeState::Experimental,               // Code state
    Tags::CPU,                             // Tags
};
const ProcessorInfo GraphToCircularPlotMesh::getProcessorInfo() const { return processorInfo_; }

GraphToCircularPlotMesh::GraphToCircularPlotMesh()
    : Processor()
    , inport_("inport")
    , networkLevelsMeshOut_("meshOut")
    , meshNetworkGraphPathNodesOut_("meshNetworkGrapshOut")
    , meshNetworkGraphPathEdgesOut_("meshNetworkGraphPathEdgesOut")
    , meshNetworkGraphPathUnifiedEdgesOut_("meshNetworkGraphPathUnifiedEdgesOut")
    , equiDistance_("equiDistance", "EquiDistance", false)
    , sortPlanarNodes_("sortPlanarNodes", "Sort Planar", false)
    , fitMeshBound_("fitMeshBound", "Fit Mesh Bound", false)
    , radialOffset_("radialOffset", "Radial Offset", 0.0f, 0.0f, 20.0f, 0.01f)
    , radialScale_("radialScale", "Radial Scale", 1.0f, 0.0f, 20.0f, 0.01f)
    , tf_("transferFunction", "Z TF")
    , tfEdgeColor_("tfEdgeColor", "Edge Color TF")
    , selectedNetworkIndex_("selectedNetworkIndex", "Network Index", 0, 2, 1)
//, networkGraphLayoutProperty_("networkGraphLayoutProperty", "NetworkGraphLayoutProperty")
{
    addPort(inport_);
    addPort(networkLevelsMeshOut_);
    addPort(meshNetworkGraphPathNodesOut_);
    addPort(meshNetworkGraphPathEdgesOut_);
    addPort(meshNetworkGraphPathUnifiedEdgesOut_);

    // addProperty(networkGraphLayoutProperty_);
    // addProperty(equiDistance_);
    addProperty(radialOffset_);
    addProperty(radialScale_);
    addProperty(sortPlanarNodes_);
    addProperty(fitMeshBound_);
    addProperty(tf_);
    addProperty(tfEdgeColor_);
    addProperty(selectedNetworkIndex_);

    inport_.onChange([&] {
        NetworkGraphInfoConstSequence networkSequence = inport_.getVectorData();
        selectedNetworkIndex_.setMaxValue(static_cast<int>(networkSequence.size()));
        selectedNetworkIndex_.setMinValue(0);
        auto val = selectedNetworkIndex_.get();
        val = std::min(val, (int)networkSequence.size() - 1);
        selectedNetworkIndex_.set(val);

        clearAllMesh();
    });

    tf_.onChange([&] { clearAllMesh(); });

    radialOffset_.onChange([&] { clearAllMesh(); });

    radialScale_.onChange([&] { clearAllMesh(); });

    selectedNetworkIndex_.onChange([&] { edgeGraphMesh_ = nullptr; });

    sortPlanarNodes_.onChange([&] {
        edgeGraphMesh_ = nullptr;
        unifiedEdgeGraphMesh_ = nullptr;
        nodeGraphMesh_ = nullptr;
    });
}

std::map<float, std::vector<LabelMetaData>> getZLevelNodeIndices(
    NetworkGraphInfoConstSequence& networkSequence, float& minPosZ, float& maxPosZ) {
    std::map<float, std::vector<LabelMetaData>> zLevelNetworks;

    for (auto& network_const : networkSequence) {
        const auto& GA = *(network_const->getAttributes());
        auto nodes = network_const->getSortedGraphNodes();

        auto firstNode = *nodes.get(0);
        auto firtstLm = extractTrajectoryLabelMetaData(network_const, firstNode);
        auto startPos = firtstLm.pos_;

        for (size_t i = 0; i < nodes.size(); i++) {
            auto n = *nodes.get(static_cast<int>(i));
            auto lm = extractTrajectoryLabelMetaData(network_const, n);
            auto pos = lm.pos_;

            auto& zlnet = zLevelNetworks[pos.z];

            bool found = false;
            for (size_t z = 0; z < zlnet.size(); z++) {
                if (zlnet[z].index_ == lm.index_) found = true;
            }

            if (!found) {
                // sort based on angle between

                float n_angle = 0.0f;

                if (glm::length2(pos - startPos) <= 0.00000001f) {
                    n_angle = 0.0f;
                } else {
                    auto v0 = vec3(1.0f, 1.0f, 0.0f);
                    auto v1 = (pos - startPos) * v0;
                    float n_dot = glm::dot(glm::normalize(v1), glm::normalize(v0));

                    n_angle = glm::acos(n_dot);
                }

                lm.weight_ = n_angle;  // theta
                zlnet.push_back(lm);
            }

            // zlnet.push_back(n);

            minPosZ = glm::min(minPosZ, pos.z);
            maxPosZ = glm::max(maxPosZ, pos.z);
        }
    }

    return zLevelNetworks;
}

std::map<int, float> getZLevelNodeDepths(NetworkGraphInfoConstSequence& networkSequence) {
    std::map<int, float> zLevelNodes;
    for (auto& network_const : networkSequence) {
        const auto& GA = *(network_const->getAttributes());
        auto nodes = network_const->getSortedGraphNodes();
        for (size_t i = 0; i < nodes.size(); i++) {
            auto n = *nodes.get(static_cast<int>(i));
            auto lm = extractTrajectoryLabelMetaData(network_const, n);
            auto pos = lm.pos_;

            zLevelNodes[lm.index_] = pos.z;
        }
    }
    return zLevelNodes;
}

void sortZLevelNetwork(std::map<float, std::vector<LabelMetaData>>& zLevelNetworks) {
    for (auto& zlevel : zLevelNetworks) {
        auto& nodes = zlevel.second;

        std::stringstream ss;
        ss << "Before sorting: ";
        for (auto& n : nodes) {
            ss << n.weight_ << ",";
        }
        ss << std::endl;
        auto theta_dcompare = [](const LabelMetaData& a, const LabelMetaData& b) {
            if (a.weight_ < b.weight_) return true;
            return false;
        };
        std::sort(nodes.begin(), nodes.end(), theta_dcompare);

        ss << "After sorting: ";
        for (auto& n : nodes) {
            ss << n.weight_ << ",";
        }
        ss << std::endl;
        ss << std::endl;
        // std::cout << ss.str();
    }
}

void GraphToCircularPlotMesh::clearAllMesh() {
    levelGraphMesh_ = nullptr;
    nodeGraphMesh_ = nullptr;
    edgeGraphMesh_ = nullptr;
    unifiedEdgeGraphMesh_ = nullptr;
}

void GraphToCircularPlotMesh::process() {
    NetworkGraphInfoConstSequence networkSequence = inport_.getVectorData();

    if (!levelGraphMesh_ || !nodeGraphMesh_ || !unifiedEdgeGraphMesh_) {
        auto newUnifiedNetwork = unifyNetwork(networkSequence);
        unifiedNetworkGraphSequence_ = NetworkGraphInfoConstSequence({newUnifiedNetwork});
    }

    if (!levelGraphMesh_) {
        levelGraphMesh_ = constructLevelsOfNetworks(unifiedNetworkGraphSequence_);
    }

    if (!nodeGraphMesh_) {
        nodeGraphMesh_ = constructCircularNodesAndEdgesOfNetworks(unifiedNetworkGraphSequence_,
                                                                  DrawType::Points, false);
    }

    if (!unifiedEdgeGraphMesh_) {
        unifiedEdgeGraphMesh_ = constructCircularNodesAndEdgesOfNetworks(
            unifiedNetworkGraphSequence_, DrawType::Lines, false);
    }

    if (!edgeGraphMesh_) {
        auto val = std::min(selectedNetworkIndex_.get(), (int)networkSequence.size() - 1);
        auto selectednetworkSequence = NetworkGraphInfoConstSequence({networkSequence[val]});
        edgeGraphMesh_ = constructCircularNodesAndEdgesOfNetworks(selectednetworkSequence,
                                                                  DrawType::Lines, true);
    }

    networkLevelsMeshOut_.setData(levelGraphMesh_);
    meshNetworkGraphPathNodesOut_.setData(nodeGraphMesh_);
    meshNetworkGraphPathEdgesOut_.setData(edgeGraphMesh_);
    meshNetworkGraphPathUnifiedEdgesOut_.setData(unifiedEdgeGraphMesh_);
}

std::shared_ptr<BasicMesh> GraphToCircularPlotMesh::constructLevelsOfNetworks(
    NetworkGraphInfoConstSequence& networkSequence) {
    float minPosZ(std::numeric_limits<float>::max());
    float maxPosZ(std::numeric_limits<float>::min());

    NetworkGraphInfoConstSequence full_networkSequence = inport_.getVectorData();

    auto zLevelNetworks = getZLevelNodeIndices(full_networkSequence, minPosZ, maxPosZ);
    // std::map<float, std::vector<int>> zLevelNetworks;
    std::map<int, float> zLevelNodes = getZLevelNodeDepths(full_networkSequence);

    // circular level lines

    auto graphCircularLevelMesh = std::make_shared<BasicMesh>();
    graphCircularLevelMesh->setModelMatrix(mat4(1.f));

    glm::vec3 center(0.0);
    auto radiusOffset = radialOffset_;
    auto segments = 60;
    auto normal = vec3(0.0f, 0.0f, 1.0f);
    auto div = (maxPosZ - minPosZ) == 0.0f ? 1.0f : (maxPosZ - minPosZ);

    for (auto& zNodes : zLevelNetworks) {
        glm::vec3 startPoint(-1.0f, 0.0f, 0.0f);

        auto radius = zNodes.first;
        auto normalizedRadius = (radius - minPosZ) / div;

        TransferFunction tf;
        glm::vec4 col = tf_.get().sample(normalizedRadius);
        tf.clearPoints();
        tf.addPoint(vec2(0.5f, 1.0f), col);

        if (radius <= 0.0f) startPoint *= -1.0f;

        radius = glm::abs(radius) + radiusOffset;

        // radius *= radialScale_;

        auto maximaCircle =
            trajectoryutils::semiCircle(center, startPoint * radius, normal, tf, segments);
        graphCircularLevelMesh->append(maximaCircle.get());

        /*auto seg = zNodes.second.size();
        auto cir = trajectoryutils::semiCircle(center, startPoint* radius, normal, tf, seg);
        graphPathMesh->append(cir.get());*/
    }

    if (fitMeshBound_) {
        {
            glm::vec3 minPos(std::numeric_limits<float>::max());
            glm::vec3 maxPos(std::numeric_limits<float>::min());
            networkgraphutil::calcualteMeshBound(graphCircularLevelMesh, minPos, maxPos);
            networkgraphutil::setModelWorldMatrix(graphCircularLevelMesh, minPos, maxPos);
        }
    }

    return graphCircularLevelMesh;
}

std::shared_ptr<BasicMesh> GraphToCircularPlotMesh::constructCircularNodesAndEdgesOfNetworks(
    NetworkGraphInfoConstSequence& networkSequence, DrawType geometryType, bool ovverrideColor) {
    NetworkGraphInfoConstSequence full_networkSequence = inport_.getVectorData();

    float minPosZ(std::numeric_limits<float>::max());
    float maxPosZ(std::numeric_limits<float>::min());

    auto zLevelNetworks = getZLevelNodeIndices(full_networkSequence, minPosZ, maxPosZ);
    if (sortPlanarNodes_) sortZLevelNetwork(zLevelNetworks);
    // std::map<float, std::vector<int>> zLevelNetworks;
    std::map<int, float> zLevelNodes = getZLevelNodeDepths(full_networkSequence);

    glm::vec3 center(0.0f);
    auto radiusOffset = radialOffset_;
    auto normal = vec3(0.0f, 0.0f, 1.0f);

    // circular points
    auto graphPathMesh = std::make_shared<BasicMesh>();
    graphPathMesh->setModelMatrix(mat4(1.f));
    auto inds = [geometryType, graphPathMesh]() {
        if (geometryType == DrawType::Points)
            return graphPathMesh->addIndexBuffer(DrawType::Points, ConnectivityType::None);
        else
            return graphPathMesh->addIndexBuffer(DrawType::Lines, ConnectivityType::None);
    }();
    std::vector<BasicMesh::Vertex> graphNodeVertices;

    // std::stringstream ss;

    for (auto& network_const : networkSequence) {
        const auto& GA = *(network_const->getAttributes());
        // auto nodes = network_const->getSortedGraphNodes();
        auto enodes = network_const->getEdgeNodes();

        /*
        auto firstNode = *nodes.begin();
        auto lastNode = *nodes.rbegin();
        auto starMaximaMd = extractTrajectoryLabelMetaData(network_const, firstNode.first);
        auto startMaximaPos = starMaximaMd.pos_;
        */

        for (size_t i = 0; i < enodes.size(); i++) {
            std::vector<ogdf::node> nodes;
            nodes.push_back(enodes[i].first);
            nodes.push_back(enodes[i].second);

            for (size_t np = 0; np < nodes.size(); np++) {
                auto nd = nodes[np];
                auto lm = extractTrajectoryLabelMetaData(network_const, nd);

                // ogdf::Color col = GA.fillColor(nd);
                // auto nodeColor = glm::vec4(col.red(), col.green(), col.blue(), col.alpha()) /
                // 256.0f;

                auto nodeColor = lm.color_;

                if (ovverrideColor) nodeColor = tfEdgeColor_.get().sample(float(i) / enodes.size());

                /*if (i == 0) {
                     ss << "StartId is :" << lm.index_ << "Color : " << lm.color_ << " Pos " <<
                lm.pos_ << std::endl;
                }
                ss << lm.index_ << " d: " << lm.pos_.z << ", ";
                */

                if (zLevelNodes.find(lm.index_) != zLevelNodes.end()) {
                    glm::vec3 startPoint(-1.0f, 0.0f, 0.0f);

                    auto posz = zLevelNodes[lm.index_];

                    auto radius = posz;  // *radialScale_;

                    // auto normalizedRadius = (radius - minPosZ) / div;
                    // glm::vec4 col = tf_.get().sample(normalizedRadius);

                    /*
                    if (nd == firstNode) {
                        nodeColor = glm::vec4(0.0f, 1.0f, 0.0f, 1.0f);
                    }
                    if (nd == lastNode) {
                        nodeColor = glm::vec4(1.0f, 0.0f, 0.0f, 1.0f);
                    }
                    */

                    /*
                    auto nit = std::find(zLevelNetworks[posz].begin(), zLevelNetworks[posz].end(),
                    lm.index_);

                    if (nit != zLevelNetworks[posz].end()) {
                        auto nDist = nit - zLevelNetworks[posz].begin();

                        auto cpoints = trajectoryutils::circlePoints(center,
                    startPoint*(glm::abs(radius) + radiusOffset), normal,
                                                                     zLevelNetworks[posz].size() +
                    1, glm::pi<float>()); auto p = cpoints[nDist+1];


                        if (radius <= 0.0) p.y *= -1.0f;

                        //nodePos.push_back(p);
                        graphNodeVertices.push_back({ p, glm::normalize(p - center), p, nodeColor
                    }); inds->add(graphNodeVertices.size());
                    }
                    else {
                        LogWarn("Segment not found");
                    }
                    */

                    for (size_t zind = 0; zind < zLevelNetworks[posz].size(); zind++) {
                        if (zLevelNetworks[posz][zind].index_ == lm.index_) {
                            auto nDist = zind;
                            auto cpoints = trajectoryutils::circlePoints(
                                center, startPoint * (glm::abs(radius) + radiusOffset), normal,
                                zLevelNetworks[posz].size() + 1, glm::pi<float>());
                            auto p = cpoints[nDist + 1];

                            // simple projection
                            /*auto p_xy = vec2(lm.pos_.x, lm.pos_.y);
                            auto s_xy = vec2(startMaximaPos.x, startMaximaPos.y);

                            auto f_xy = (p_xy - s_xy);

                            auto p = vec3(f_xy, center.z);*/

                            if (radius <= 0.0) p.y *= -1.0f;

                            // nodePos.push_back(p);
                            graphNodeVertices.push_back(
                                {p, glm::normalize(p - center), p, nodeColor});
                            inds->add(static_cast<uint32_t>(graphNodeVertices.size()));
                            break;
                        }
                    }
                } else {
                    LogWarn("Level Node not found");
                }
            }
        }
    }
    graphNodeVertices.push_back(graphNodeVertices.back());
    graphPathMesh->addVertices(graphNodeVertices);

    // LogWarn(ss.str());

    if (fitMeshBound_) {
        {
            glm::vec3 minPos(std::numeric_limits<float>::max());
            glm::vec3 maxPos(std::numeric_limits<float>::min());
            networkgraphutil::calcualteMeshBound(graphPathMesh, minPos, maxPos);
            networkgraphutil::setModelWorldMatrix(graphPathMesh, minPos, maxPos);
        }
    }

    return graphPathMesh;
}

}  // namespace inviwo
