/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2016-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <modules/morphologyvisualization/processors/skeletonextraction.h>
#include "inviwo/core/util/volumeramutils.h"
#include <inviwo/core/util/indexmapper.h>
#include <modules/morphologyvisualization/datastructures/skeletonvoxels.h>
#include <modules/morphologyvisualization/util/skeletonadjacencymapper.h>
#include <modules/base/datastructures/kdtree.h>
#include <queue>

namespace inviwo {

using namespace skeletonutil;



// The Class Identifier has to be globally unique. Use a reverse DNS naming scheme
const ProcessorInfo SkeletonExtraction::processorInfo_{
    "org.inviwo.SkeletonExtraction",  // Class identifier
    "SkeletonExtraction",             // Display name
    "Morphology",                     // Category
    CodeState::Experimental,          // Code state
    Tags::None,                       // Tags
};
const ProcessorInfo SkeletonExtraction::getProcessorInfo() const { return processorInfo_; }

SkeletonExtraction::SkeletonExtraction()
    : Processor()
    , binaryVolume_("binaryVolume")
    , distanceVolume_("distanceVolume")
    , skeletalPointsMeshOutport_("outport")
    , distanceThreshold_("distanceThreshold", "Distance Threshold", 0.98f, 0.0f, 1.0f, 0.001f)
    , extractionMethod_("extractionMethod", "Extraction Method")
    , visibilityProperties_("visibilityProperties","Visibility")
    , context_("contex", "Contex")
    , showBackGround_("showBackGround", "Show BackGround", false)
    , showExteriors_("showExteriors", "Show Exterior", true)
    , showEndPoints_("showEndPoints", "Show EndPoints", false)
    , showMaximalPoints_("showMaximalPoints", "Show MaximalPoints", false)
    , showSkeleton_("showSkeleton", "Show Skeleton", false)
    , contextOptions_("contextOptions", "Context Options")
    //, showForeGround_("showForeGround", "Show ForeGround", true) 
    //, showSegments_("showSegments", "Show Segments", false)
    //, showDistanceTransform_("showDistanceTransform", "Show DistanceTransform", false) 
    , itreations_("thinningItreations", "Thinning Itreations", 1, 0, 100, 1)
    {
    addPort(binaryVolume_);
    addPort(distanceVolume_);
    addPort(skeletalPointsMeshOutport_);

    addProperty(extractionMethod_);
    addProperty(distanceThreshold_);

    
    visibilityProperties_.addProperty(showBackGround_);
    visibilityProperties_.addProperty(showExteriors_);
    visibilityProperties_.addProperty(showEndPoints_);	
    visibilityProperties_.addProperty(showMaximalPoints_);
    visibilityProperties_.addProperty(showSkeleton_);
    //visibilityProperties_.addProperty(showForeGround_); 
    //visibilityProperties_.addProperty(showSegments_); 
    //visibilityProperties_.addProperty(showDistanceTransform_);
    addProperty(visibilityProperties_);

    
    context_.addProperty(contextOptions_);
    contextOptions_.addOption("foreGroundOnly", "ForeGroundOnly", ContextVoxelVisibility::ForeGroundOnly);
    contextOptions_.addOption("segmentation", "Segmentation", ContextVoxelVisibility::Segments);
    contextOptions_.addOption("distanceTransform", "Distance Transform", ContextVoxelVisibility::DistanceTransform);
    contextOptions_.setCurrentStateAsDefault();
    addProperty(context_);

    addProperty(itreations_);


    distanceThreshold_.onChange([&] { skeletalMeshes_.clear(); });
    context_.onChange([&] { skeletalMeshes_.clear(); });
    extractionMethod_.onChange([&] { 
        skeletalMeshes_.clear(); 
        if (extractionMethod_.get() == ExtractionMode::DOHThinning) {
            itreations_.setVisible(true);
        }
        else itreations_.setVisible(false);		
    });

    extractionMethod_.addOption("thresholdFilterValue", "ThresholdFilterValue",
                                (int)ExtractionMode::ThresholdFilterValue);
    extractionMethod_.addOption("dohthinning", "DOHThinning", (int)ExtractionMode::DOHThinning);
    extractionMethod_.setCurrentStateAsDefault();

    distanceVolume_.onChange([&] { if (ski_!=nullptr) ski_->reComputeDistance_ = true; });
}



void SkeletonExtraction::process() {
    // outport_.setData(myImage);

    if (!skeletalMeshes_.size()) {
        extractSkeleton();
    }

    skeletalMeshes_.clear();
    auto fskeleton = generateVoxelMesh(ski_->foregroundVoxels_);
    

    if (showBackGround_) {
        auto bskeleton = generateVoxelMesh(ski_->backgroundVoxels_);
        fskeleton->append(bskeleton.get());
    }

    skeletalMeshes_.push_back(fskeleton);

    skeletalPointsMeshOutport_.setData(skeletalMeshes_.front());
}

void SkeletonExtraction::fillupVoxelInfo(VoxelMap& allVoxels, VoxelMap& foregroundVoxels,
                                      VoxelMap& backgroundVoxels) {

    std::function<void(const size3_t&)> func = [&](const size3_t& pos) {};

    // construct voxels
    func = [&](const size3_t& pos) {

        auto cpos = glm::clamp(ivec3(pos), ivec3(0), ivec3(ski_->dim_) - ivec3(1));
        auto& tree = ski_->tree_;

        // Voxel flags: exterior, queued, deletable

        ivwAssert(allVoxels.find(ski_->index_(cpos)) == allVoxels.end(), "Voxel entry already exist");

        // allVoxels[index(cpos)] = Voxel();
        // Voxel& v = allVoxels[index(cpos)];
        Voxel v;
        v.init();
        v.pos_ = cpos;
        v.index_ = ski_->index_(cpos);
        v.distance_ = (ski_->distVolume_->getRepresentation<VolumeRAM>()->getAsNormalizedDouble(pos));
        v.value_ = (ski_->maskVolume_->getRepresentation<VolumeRAM>()->getAsNormalizedDouble(pos));
        v.neighbours_ = 0;
        v.flags_ = 0;
        v.labels_ = 0;
        v.segmentationValue_ = 0.0;
        v.mapping_ = ski_->voxelMapping_;
        v.boundaryTest_ = ski_->voxelBoundaryTest_;
        v.getNeighbours_ = ski_->voxelNeighbourHood_;
        v.getNeighbourVoxels_ = ski_->voxelNeighbourHoodVoxels_;
        v.getNormalizedVoxelPos_ = ski_->voxelgetNormalizedPos_;

        for (int x = 0; x < 3; x++)
            for (int y = 0; y < 3; y++)
                for (int z = 0; z < 3; z++) v.unsetNeighbour(ivec3(x, y, z));

        v.setVoxelFlag(VoxelFlags::None);

        if (ski_->isForeGround_(cpos))
            v.setVoxelFlag(VoxelFlags::ForeGround);
        else
            v.setVoxelFlag(VoxelFlags::BackGround);

        allVoxels[ski_->index_(cpos)] = v;

        ski_->tree_.insert(cpos, v);

    };
    util::forEachVoxel(*ski_->maskVolume_->getRepresentation<VolumeRAM>(), func);

    skeletonutil::getFlagedVoxels(allVoxels, foregroundVoxels, VoxelFlags::ForeGround); //foregrounds
    skeletonutil::getFlagedVoxels(allVoxels, backgroundVoxels, VoxelFlags::BackGround); //backgrounds


    for (auto& ov : allVoxels) skeletonutil::setNeighbourInfo(allVoxels, ov.second, ski_->dim_);

    for (auto& fv : foregroundVoxels) skeletonutil::setNeighbourInfo(allVoxels, fv.second, ski_->dim_);

    for (auto& bv : backgroundVoxels) skeletonutil::setNeighbourInfo(allVoxels, bv.second, ski_->dim_);

}


void SkeletonExtraction::extractSkeleton() {
    if (!binaryVolume_.getData() || !distanceVolume_.getData()) {
        return;
    }   

    
    size3_t dim = binaryVolume_.getData()->getDimensions();
    ivec3 rdim(ivec3(dim) - ivec3(1));

    auto recomDist = true;

    if (ski_ != nullptr)
        recomDist = ski_->reComputeDistance_;
    else
        ski_ = std::make_shared<SkeletonInfo>(binaryVolume_.getData(), distanceVolume_.getData());

    VoxelMap& allVoxels = ski_->allVoxels_;
    VoxelMap& foregroundVoxels = ski_->foregroundVoxels_;
    VoxelMap& backgroundVoxels = ski_->backgroundVoxels_;

    if (recomDist) {
        /*ski_->dim_ = dim;
        ski_->rdim_ = rdim;
        ski_->binaryVolume_ = binaryVolume_.getData();
        ski_->distVolume_ = distanceVolume_.getData();*/
        ski_->tree_.clear();
        ski_->thres_ = distanceThreshold_.get();
        ski_->reComputeDistance_ = recomDist;

        allVoxels = ski_->allVoxels_;
        foregroundVoxels = ski_->foregroundVoxels_;
        backgroundVoxels = ski_->backgroundVoxels_;		

        if (ski_->reComputeDistance_) {
            {
                IVW_CPU_PROFILING("Building Voxels.");
                ski_->fillupVoxelInfo();
                skeletonutil::getFlagedVoxels(allVoxels, foregroundVoxels, VoxelFlags::ForeGround); //foregrounds
                skeletonutil::getFlagedVoxels(allVoxels, backgroundVoxels, VoxelFlags::BackGround); //backgrounds


                for (auto& ov : allVoxels) skeletonutil::setNeighbourInfo(allVoxels, ov.second, ski_->dim_);
                for (auto& fv : foregroundVoxels) skeletonutil::setNeighbourInfo(allVoxels, fv.second, ski_->dim_);
                for (auto& bv : backgroundVoxels) skeletonutil::setNeighbourInfo(allVoxels, bv.second, ski_->dim_);
            }

            {
                IVW_CPU_PROFILING("Compute DistanceTransform");
                ski_->minMaxDist_ = computeDistanceTransform(backgroundVoxels, foregroundVoxels);
                ski_->reComputeDistance_ = false;
            }
        }

    }

    


    if (extractionMethod_.get() == ExtractionMode::ThresholdFilterValue) {
        
    }

    if (extractionMethod_.get() == ExtractionMode::DOHThinning) {
        // Required basic functions
        // isDeletable()
        //  isSimple() - Rule A
        //  isEndPointPointOfMedialEdge() - Rule A
        //  isCenterOfMaximalSphere() - Rule B

        // isExterior()
        //  Let (m,n) is (6, 26)
        //  isN6Adjacent()==true then exterior

        // Main algorithm
        // 1. For each foreground voxel
        // Voxel flags: exterior, queued, deletable
        // 1. Gather all exterior points
        //  a. queue exterior points, flag them as queued
        // 2. mark foregrounds and backgrounds points

        // 2. Process exterior points until there are none (LOOP)
        //  a. if isDeletable()
        //       mark as deleted (that is set it as background)
        //         for all N26 neighbors
        //           if foreground and not queued
        //             queue them
        //    else
        //      flag them as non-deletable
        

        NAdjacencyMapper np;
        for (auto & f : foregroundVoxels) {
            auto& v = f.second;

            if (np.isExterior(v)) {
                v.setVoxelFlag(VoxelFlags::Exterior);
            }
            else {
                v.setVoxelFlag(VoxelFlags::Interior);
            }

            if (np.isEndPoint(v)) {
                v.setVoxelFlag(VoxelFlags::EndPoint);
            }

            v.unsetVoxelFlag(VoxelFlags::Queued);
        }
        
        

        //normalize distance
        //for (auto& fv : foregroundVoxels)
        //	fv.second.distance_ = (fv.second.distance_ - minMaxDist.x) / (minMaxDist.y - minMaxDist.x);		

        auto minMaxDist = ski_->minMaxDist_;

        if (showMaximalPoints_ || showSkeleton_) {

            IVW_CPU_PROFILING("Compute MaximalPoint");
            
            auto maximalCenters = getMaximalCenters(foregroundVoxels, glm::length(vec3(1.1f))); // minMaxDist.y - minMaxDist.x)*ski_->thres_);

            for (auto& mv : maximalCenters) {
                foregroundVoxels[mv.index_].setVoxelFlag(VoxelFlags::Maximum);
            }

            LogInfo("Maximal Center Count: " << maximalCenters.size());
        }


        if (showSkeleton_) {
                        
            //auto foregroundCopy = foregroundVoxels;
            IVW_CPU_PROFILING("Compute Skeleton");

            //fill voxel queue
            auto cmp = [](const Voxel& left, const Voxel& right) { return (left.distance_) < (right.distance_); };

            for (auto& wv : foregroundVoxels) {
                if (np.isExterior(wv.second)) {
                    wv.second.setVoxelFlag(VoxelFlags::Exterior);
                }
                else {
                    wv.second.unsetVoxelFlag(VoxelFlags::Exterior);
                }

                if (wv.second.hasVoxelFlag(VoxelFlags::ForeGround) &&
                    wv.second.hasVoxelFlag(VoxelFlags::Exterior)) {
                    wv.second.setVoxelFlag(VoxelFlags::Queued);
                }
            }
            
            for (int it = 0; it <= itreations_.get(); it++) {

                std::priority_queue<Voxel, std::vector<Voxel>, decltype(cmp)> voxelQueue(cmp);

                for (auto& wv : foregroundVoxels)
                    skeletonutil::setNeighbourInfo(foregroundVoxels, wv.second, ski_->dim_); //foregrounds

                for (auto& wv : foregroundVoxels) {
                    if (wv.second.hasVoxelFlag(VoxelFlags::ForeGround) &&
                        wv.second.hasVoxelFlag(VoxelFlags::Queued)) {
                        voxelQueue.push(wv.second);
                    }
                }

                std::cout << "Totoal Foregroundsr Voxels : " << foregroundVoxels.size() << std::endl;
                std::cout << "Current Exterior Voxels in Queue are: " << voxelQueue.size() << std::endl;
                                
                while (voxelQueue.size()) {
                    auto qv = voxelQueue.top();
                    skeletonutil::setNeighbourInfo(foregroundVoxels, qv, ski_->dim_); //foregrounds

                    if (!foregroundVoxels[qv.index_].hasVoxelFlag(VoxelFlags::Maximum))
                    {

                        //if deletable meaning ( simple + !endpoint)
                        if (np.isDeletable(qv)) 
                        {
                            auto nghs = foregroundVoxels[qv.index_].getNeighbourVoxels_(qv.pos_, 1.5);

                            foregroundVoxels[qv.index_].unsetVoxelFlag(VoxelFlags::ForeGround);
                            foregroundVoxels[qv.index_].setVoxelFlag(VoxelFlags::BackGround);

                            //neighbour foregrounds need to be queued
                            
                            for (auto& ngv: nghs)
                                foregroundVoxels[ngv.index_].setVoxelFlag(VoxelFlags::Queued);

                            //foregroundVoxels.erase(qv.index_);
                        }
                    }
                    voxelQueue.pop();
                }

                //voxelQueue = ith_voxelQueue;
                //break;
            }

            VoxelMap tempList;
            skeletonutil::getFlagedVoxels(foregroundVoxels, tempList, VoxelFlags::ForeGround); //foregrounds

            for (auto& wv : foregroundVoxels) wv.second.setVoxelFlag(VoxelFlags::BackGround);

            for (auto& wv : foregroundVoxels)
                skeletonutil::setNeighbourInfo(foregroundVoxels, wv.second, ski_->dim_); //foregrounds			

            for (auto& wv : foregroundVoxels) {
                if (np.isExterior(wv.second)) {
                    wv.second.setVoxelFlag(VoxelFlags::Exterior);
                }
                else {
                    wv.second.unsetVoxelFlag(VoxelFlags::Exterior);
                }
            }


            for (auto& tv : tempList) {
                foregroundVoxels[tv.second.index_] = tv.second;
                foregroundVoxels[tv.second.index_].setVoxelFlag(VoxelFlags::Skeleton);
                foregroundVoxels[tv.second.index_].setVoxelFlag(VoxelFlags::ForeGround);
            }
            
        }

        if (contextOptions_.get() == ContextVoxelVisibility::Segments) {
            
            auto segments = getLabeledVoxelSegments(foregroundVoxels, ski_->thres_*(minMaxDist.x - minMaxDist.y));

            auto segfactor = 1.0f / segments.size();

            int segCount = 0;
            for (auto& seg : segments) {
                auto s = segfactor*segCount;
                for (auto& sv : seg.second) {
                    foregroundVoxels[sv.index_].segmentationValue_ = s;
                }
                segCount++;
            }

            LogInfo("Foreground Voxel Segment Count: " << segments.size());
        }
        else if (contextOptions_.get() == ContextVoxelVisibility::DistanceTransform) {
            //normalize
            for (auto & sv : foregroundVoxels) {
                auto nd = (foregroundVoxels[sv.second.index_].distance_ - minMaxDist.x) / (minMaxDist.y - minMaxDist.x);
                nd = glm::clamp(nd, 0.0, 1.0);
                foregroundVoxels[sv.second.index_].segmentationValue_ = nd;
            }
        }
        else if (contextOptions_.get() == ContextVoxelVisibility::ForeGroundOnly) {
            for (auto & sv : foregroundVoxels)
                foregroundVoxels[sv.second.index_].segmentationValue_ = 0.5f;
        }
    }
    
}

std::shared_ptr<BasicMesh> SkeletonExtraction::generateVoxelMesh(const VoxelMap& voxelMap) {

    NAdjacencyMapper np;
    
    auto dim = ski_->dim_;
    std::shared_ptr<BasicMesh> skeleton = std::make_shared<BasicMesh>();
    skeleton->setModelMatrix(ski_->maskVolume_->getModelMatrix());
    skeleton->setWorldMatrix(ski_->maskVolume_->getWorldMatrix());

    auto ind = skeleton->addIndexBuffer(DrawType::Points, ConnectivityType::None);
    std::vector<BasicMesh::Vertex> vertices;
    vertices.reserve(dim.x *dim.y * dim.z);


    TransferFunction tf((int)512);
    tf.clearPoints();
    tf.addPoint(vec2(0.0f, 1.0f), vec4(1.0f, 0.0f, 0.0f, 1.0f)); //red 
    tf.addPoint(vec2(0.5f, 1.0f), vec4(1.0f, 1.0f, 1.0f, 1.0f)); //white
    tf.addPoint(vec2(1.0f, 1.0f), vec4(0.0f, 0.0f, 1.0f, 1.0f)); //blue

    
    std::function<void(const VoxelMap& srcVoxelMap)> computeVertices = [&](const VoxelMap& srcVoxelMap) {

        VoxelMap voxels = srcVoxelMap;

        while (voxels.size()) {

            auto f = voxels.begin();
            const auto& v = f->second;
            auto pos = v.pos_;

            
            ind->add(static_cast<uint32_t>(vertices.size()));

            auto color = vec4(1.0f, 0.1f, 0.1f, 1.0f);

            if (v.hasVoxelFlag(VoxelFlags::ForeGround)) {
                color = tf.sample(v.segmentationValue_);

                if (np.isExterior(v) && showExteriors_) {
                    color = vec4(0.8f, 0.8f, 0.1f, 1.0f);
                }
                else if (np.isEndPoint(v) && showEndPoints_) {
                    color = vec4(1.0f, 1.0f, 0.1f, 1.0f);
                }
                else if (v.hasVoxelFlag(VoxelFlags::Maximum) && showMaximalPoints_) {
                    color = vec4(0.1f, 0.8f, 0.1f, 1.0f);
                }
                else if (v.hasVoxelFlag(VoxelFlags::Skeleton) && showSkeleton_) {
                    color = vec4(0.1f, 0.1f, 0.1f, 1.0f);
                }
                /*else {
                    voxels.erase(f);
                    continue;
                }*/

                color.a = 1.0f;
                vertices.push_back({ vec3(pos) / vec3(dim), vec3(pos) / vec3(dim),
                    vec3(pos) / vec3(dim), color });

            }

            if (v.hasVoxelFlag(VoxelFlags::BackGround)) {
                color = vec4(0.35f, 0.35f, 0.35f, 1.0f);

                if (!showBackGround_) {
                    voxels.erase(f);
                    continue;
                }

                color.a = 1.0f;
                vertices.push_back({ vec3(pos) / vec3(dim), vec3(pos) / vec3(dim),
                    vec3(pos) / vec3(dim), color });
            }

            voxels.erase(f);
        };
    };

    //if (showForeGround_) 
    {
        computeVertices(voxelMap);
    }
    
    //Final result


    if (vertices.size())
        skeleton->addVertices(vertices);

    return skeleton;
}

}  // namespace
