/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2017-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <warn/push>
#include <warn/ignore/all>
#include <gtest/gtest.h>
#include <warn/pop>

#include <inviwo/core/common/inviwo.h>

#include <inviwo/core/datastructures/volume/volume.h>
#include <inviwo/core/datastructures/volume/volumeramprecision.h>
#include <inviwo/core/util/indexmapper.h>

#include <modules/morphologyvisualization/util/batchvolumesampler.h>

#include <vector>
#include <glm/gtc/epsilon.hpp>

//#pragma optimize("", off)

namespace inviwo {

TEST(BatchVolumeSampler, PBCTest) {
    const auto dim = size3_t(3, 3, 3);

    auto vr = std::make_shared<VolumeRAMPrecision<float>>(dim);
    auto data = vr->getDataTyped();
    auto volume = std::make_shared<Volume>(vr);

    util::IndexMapper3D im(dim);

    data[im(0, 0, 0)] = 4.0f;
    data[im(1, 0, 0)] = 2.0f;
    data[im(2, 0, 0)] = 8.0f;
    data[im(0, 1, 0)] = 0.0f;
    data[im(1, 1, 0)] = 0;
    data[im(2, 1, 0)] = 0;
    data[im(0, 2, 0)] = 16.0f;
    data[im(1, 2, 0)] = 0;
    data[im(2, 2, 0)] = 0;

    data[im(0, 0, 1)] = 0;
    data[im(1, 0, 1)] = 0;
    data[im(2, 0, 1)] = 0;
    data[im(0, 1, 1)] = 0;
    data[im(1, 1, 1)] = 0;
    data[im(2, 1, 1)] = 0;
    data[im(0, 2, 1)] = 0;
    data[im(1, 2, 1)] = 0;
    data[im(2, 2, 1)] = 0;

    data[im(0, 0, 2)] = 0;
    data[im(1, 0, 2)] = 0;
    data[im(2, 0, 2)] = 0;
    data[im(0, 1, 2)] = 0;
    data[im(1, 1, 2)] = 0;
    data[im(2, 1, 2)] = 0;
    data[im(0, 2, 2)] = 0;
    data[im(1, 2, 2)] = 0;
    data[im(2, 2, 2)] = 9.0f;

    {
        const auto voxelDist = 1.0 / static_cast<double>(dim.x);
        const auto half = 0.5 * voxelDist;

        std::vector<dvec3> samplePositions = {
            {0, 0, 0},                                      // -> 4.0f
            {2 * voxelDist, 2 * voxelDist, 2 * voxelDist},  // ->  9.0
            {1, 1, 1},                                      // -> 9.0f
            {voxelDist, 0, 0},                              // --> 2.0f
            {-voxelDist, 0, 0},                             // --> 8.0f (repeat) / 4.0 (clamp)
            {voxelDist, voxelDist, voxelDist},              // -> 0.0f
            {half, 0, 0},                                   // -> 3.0 (linear) / 4.0 (nearest)
            {-half, 0, 0},  // --> 6.0 (repeat linear) / 9.0 (repeat nearest) / 4.0 (clamp)
        };

        {
            std::vector<double> expectedClampNearest = {
                4.0, 9.0, 9.0, 2.0, 4.0, 0.0, 4.0, 4.0,
            };
            std::vector<double> result(samplePositions.size());
            auto sampler =
                util::makeBatchVolumeSampler<1, SampleBehaviour::Nearest, BoundaryBehaviour::Clamp,
                                             BoundaryBehaviour::Clamp, BoundaryBehaviour::Clamp>(
                    volume);
            sampler->sample(samplePositions, result);

            EXPECT_EQ(expectedClampNearest, result) << "Clamp Nearest";
        }

        {
            std::vector<double> expectedClamp = {
                4.0, 9.0, 9.0, 2.0, 4.0, 0.0, 3.0, 4.0,
            };
            std::vector<double> result(samplePositions.size());
            auto sampler =
                util::makeBatchVolumeSampler<1, SampleBehaviour::Trilinear,
                                             BoundaryBehaviour::Clamp, BoundaryBehaviour::Clamp,
                                             BoundaryBehaviour::Clamp>(volume);
            sampler->sample(samplePositions, result);

            for (std::size_t i = 0; i < result.size(); ++i) {
                EXPECT_DOUBLE_EQ(expectedClamp[i], result[i])
                    << "Input i = " << i << ": " << samplePositions[i] << " (clamp trilinear)";
            }
        }

        {
            std::vector<double> expectedRepeat = {
                4.0, 9.0, 4.0, 2.0, 8.0, 0.0, 3.0, 6.0,
            };
            std::vector<double> result(samplePositions.size());
            auto sampler =
                util::makeBatchVolumeSampler<1, SampleBehaviour::Trilinear,
                                             BoundaryBehaviour::Repeat, BoundaryBehaviour::Repeat,
                                             BoundaryBehaviour::Repeat>(volume);
            sampler->sample(samplePositions, result);

            for (std::size_t i = 0; i < result.size(); ++i) {
                EXPECT_DOUBLE_EQ(expectedRepeat[i], result[i])
                    << "Input i = " << i << ": " << samplePositions[i] << " (repeat trilinear)";
            }
        }
    }
}

//#pragma optimize("", on)

}  // namespace
