
/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2014-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#include <warn/push>
#include <warn/ignore/all>
#include <gtest/gtest.h>
#include <warn/pop>

#include <inviwo/core/util/indexmapper.h>
#include <modules/morphologyvisualization/datastructures/skeletonvoxels.h>
#include <modules/morphologyvisualization/util/skeletonadjacencymapper.h>
#include <string>
#include <inviwo/core/datastructures/volume/volume.h>
#include <inviwo/core/datastructures/volume/volumeram.h>
#include <inviwo/core/util/volumeramutils.h>
#include <modules/base/datastructures/kdtree.h>


using namespace inviwo::skeletonutil;

namespace inviwo {

void fillNeighbours(skeletonutil::Voxel& v, std::vector<int>& vpoints, skeletonutil::NAdjacencyMapper& nm) {

	v.neighbours_ = 0;
	v.flags_ = 0;
	v.setVoxelFlag(skeletonutil::VoxelFlags::ForeGround);


	auto ind = nm.getN27(ivec3(1));
	for (auto& i : ind) {
		if (vpoints[nm.mapper3x3_(i)] == 1)
			v.setNeighbour(size3_t(i));
		else
			v.unsetNeighbour(size3_t(i));
	}
}

int countVoxelValues(std::vector<int>& vpoints, int value) {
	auto count = 0;

	for (size_t i = 0; i < vpoints.size(); i++)
		if (vpoints[i] == value) count++;

	return count;
}

void generateTest3x3VoxelPoints_allblack(std::vector<int>& vpoints, int& expectedLabelCount) {
	/*
	+-----------+
	| 0 | 0 | 0 |
	+-----------+
	| 0 | 1 | 0 |
	+-----------+
	| 0 | 0 | 0 |
	+---------- +
	*/

	util::IndexMapper3D mapper(size3_t(3));

	for (size_t i = 0; i < 27; i++)
		vpoints.push_back(0);

	vpoints[mapper(size3_t(1, 1, 1))] = 1;

	expectedLabelCount = 0;
}

void generateTest3x3VoxelPoints_allwhite(std::vector<int>& vpoints, int& expectedLabelCount) {
	/*
	+-----------+
	| 1 | 1 | 1 |
	+-----------+
	| 1 | 1 | 1 |
	+-----------+
	| 1 | 1 | 1 |
	+---------- +
	*/
	util::IndexMapper3D mapper(size3_t(3));

	for (size_t i = 0; i < 27; i++)
		vpoints.push_back(1);
	expectedLabelCount = 1;
}
void generateTest3x3VoxelPoints_Case0(std::vector<int>& vpoints, int& expectedLabelCount) {
	/*
	+-----------+
	| 0 | 0 | 0 |
	+-----------+
	| 0 | 1 | 0 |
	+-----------+
	| 0 | 1 | 0 |
	+---------- +
	*/

	util::IndexMapper3D mapper(size3_t(3));

	for (size_t i = 0; i < 27; i++)
		vpoints.push_back(0);

	vpoints[mapper(size3_t(1, 0, 1))] = 1;
	vpoints[mapper(size3_t(1, 1, 1))] = 1;

	expectedLabelCount = 1;
}

void generateTest3x3VoxelPoints_Case1(std::vector<int>& vpoints, int& expectedLabelCount) {
	/*
		+-----------+
		| 1 | 1 | 1 |
		+-----------+
		| 0 | 1 | 0 |
		+-----------+
		| 1 | 1 | 1 |
		+---------- +
	*/

	util::IndexMapper3D mapper(size3_t(3));

	for (size_t i = 0; i < 27; i++)
		vpoints.push_back(0);

	vpoints[mapper(size3_t(0, 0, 1))] = 1;
	vpoints[mapper(size3_t(1, 0, 1))] = 1;
	vpoints[mapper(size3_t(2, 0, 1))] = 1;

	vpoints[mapper(size3_t(1, 1, 1))] = 1;

	vpoints[mapper(size3_t(0, 2, 1))] = 1;
	vpoints[mapper(size3_t(1, 2, 1))] = 1;
	vpoints[mapper(size3_t(2, 2, 1))] = 1;	

	expectedLabelCount = 2;
}



void generateTest3x3VoxelPoints_Case2(std::vector<int>& vpoints, int& expectedLabelCount) {
	/*
	+-----------+
	| 1 | 1 | 1 |
	+-----------+
	| 0 | 1 | 1 |
	+-----------+
	| 1 | 1 | 1 |
	+---------- +
	*/

	util::IndexMapper3D mapper(size3_t(3));

	for (size_t i = 0; i < 27; i++)
		vpoints.push_back(0);

	vpoints[mapper(size3_t(0, 0, 1))] = 1;
	vpoints[mapper(size3_t(1, 0, 1))] = 1;
	vpoints[mapper(size3_t(2, 0, 1))] = 1;

	vpoints[mapper(size3_t(1, 1, 1))] = 1;
	vpoints[mapper(size3_t(2, 1, 1))] = 1;

	vpoints[mapper(size3_t(0, 2, 1))] = 1;
	vpoints[mapper(size3_t(1, 2, 1))] = 1;
	vpoints[mapper(size3_t(2, 2, 1))] = 1;

	expectedLabelCount = 1;
}

void generateTest3x3VoxelPoints_Case3(std::vector<int>& vpoints, int& expectedLabelCount) {
	/*
	+-----------+
	| 1 | 1 | 1 |
	+-----------+
	| 1 | 1 | 1 |
	+-----------+
	| 1 | 1 | 1 |
	+---------- +
	*/

	util::IndexMapper3D mapper(size3_t(3));

	for (size_t i = 0; i < 27; i++)
		vpoints.push_back(0);

	vpoints[mapper(size3_t(0, 0, 1))] = 1;
	vpoints[mapper(size3_t(1, 0, 1))] = 1;
	vpoints[mapper(size3_t(2, 0, 1))] = 1;

	vpoints[mapper(size3_t(0, 2, 1))] = 1;
	vpoints[mapper(size3_t(1, 2, 1))] = 1;
	vpoints[mapper(size3_t(2, 2, 1))] = 1;

	vpoints[mapper(size3_t(0, 1, 1))] = 1;
	vpoints[mapper(size3_t(1, 1, 1))] = 1;
	vpoints[mapper(size3_t(2, 1, 1))] = 1;

	expectedLabelCount = 1;
}

void runStandardTests(skeletonutil::Voxel& v, std::vector<int>& vpoints, skeletonutil::NAdjacencyMapper& nm, std::string testName="") {

	//////////////////////////////////////////////////////////////////////////
	//test foreground count	
	auto expectedValue = countVoxelValues(vpoints, 1);
	auto obtainedValue = nm.countN_26(v) + 1;
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " Foreground count failed";


	//////////////////////////////////////////////////////////////////////////
	//test neighbour flags
	std::vector<size_t> input;
	auto n27i = nm.getN27(ivec3(1));
	for (size_t i = 0; i < n27i.size(); i++) input.push_back(0);

	// We use (m,n) = (26,6)
	for (auto& n : n27i) {
		auto p = v.hasNeighbour(n);
		input[nm.mapper3x3_(n)] = p;
	}
	expectedValue = 27;
	obtainedValue = 0;
	for (size_t i = 0; i < n27i.size(); i++) {
		if (input[i] == vpoints[i])
			obtainedValue++;
	}
	EXPECT_EQ(expectedValue, obtainedValue) << testName << "Neighbour flagging failed";
}

TEST(MorphologySkeletonizationTests, VoxelLabelTest) {

	std::vector<int> vpoints;
	
	skeletonutil::Voxel v;
	skeletonutil::NAdjacencyMapper nm;

	auto expectedValue = 0;
	auto obtainedValue = 0;
	auto testCount = 0;
	
	v.pos_ = ivec3(1);
	v.index_ = nm.mapper3x3_(ivec3(1));
	v.distance_ = 1.0f;
	v.value_ = 1.0f;
	v.segmentationValue_ = 0;
	v.neighbours_ = 0;
	v.flags_ = 0;
	v.labels_ = 0;
	v.setVoxelFlag(skeletonutil::VoxelFlags::ForeGround);
		
	std::string testName;
	//////////////////////////////////////////////////////////////////////////
	//test all black voxels
	testName = "Test Case All Black - ";
	vpoints.clear();
	generateTest3x3VoxelPoints_allblack(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);
	obtainedValue = nm.countComponents_N26(v);
	EXPECT_EQ(expectedValue, obtainedValue) << testName <<  " Label Count Failed";

	runStandardTests(v, vpoints, nm, testName);


	//////////////////////////////////////////////////////////////////////////
	//test all white voxels
	testName = "Test Case All White - ";
	vpoints.clear();
	generateTest3x3VoxelPoints_allwhite(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);
	obtainedValue = nm.countComponents_N26(v);
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " Label Count Failed";

	runStandardTests(v, vpoints, nm, testName);
	testCount++;
	
	//////////////////////////////////////////////////////////////////////////
	//test label count - case1
	testName = "Test Case - " + toString(testCount);
	vpoints.clear();
	generateTest3x3VoxelPoints_Case1(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);
	obtainedValue = nm.countComponents_N26(v);
	EXPECT_EQ(expectedValue, obtainedValue) << testName  << " Label Count Failed";

	runStandardTests(v, vpoints, nm, testName);
	testCount++;

	//////////////////////////////////////////////////////////////////////////
	//test label count - case2
	testName = "Test Case - " + toString(testCount);
	vpoints.clear();
	generateTest3x3VoxelPoints_Case2(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);
	obtainedValue = nm.countComponents_N26(v);
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " Label Count Failed";
		
	runStandardTests(v, vpoints, nm, testName);
	testCount++;


	//////////////////////////////////////////////////////////////////////////
	//test label count - case3
	testName = "Test Case - " + toString(testCount);
	vpoints.clear();
	generateTest3x3VoxelPoints_Case3(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);
	obtainedValue = nm.countComponents_N26(v);
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " Label Count Failed";

	runStandardTests(v, vpoints, nm, testName);
	testCount++;

	//////////////////////////////////////////////////////////////////////////
	//test exterior points
	testName = "Test Exterior Points - " + toString(testCount);
	vpoints.clear();
	generateTest3x3VoxelPoints_Case1(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);
	obtainedValue = nm.countComponents_N26(v);
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " Label Count Failed";

	runStandardTests(v, vpoints, nm, testName);
	testCount++;
	
	
}


TEST(MorphologySkeletonizationTests,ExteriorPointTest) {

	std::vector<int> vpoints;

	skeletonutil::Voxel v;
	skeletonutil::NAdjacencyMapper nm;

	auto expectedValue = 0;
	auto obtainedValue = 0;
	auto testCount = 0;

	v.pos_ = ivec3(1);
	v.index_ = nm.mapper3x3_(ivec3(1));
	v.distance_ = 1.0f;
	v.value_ = 1.0f;
	v.segmentationValue_ = 0;
	v.neighbours_ = 0;
	v.flags_ = 0;
	v.labels_ = 0;
	v.setVoxelFlag(skeletonutil::VoxelFlags::ForeGround);

	std::string testName;	

	//////////////////////////////////////////////////////////////////////////
	//test exterior points - all white
	testName = "Test Exterior Points All white - " + toString(testCount);
	vpoints.clear();
	generateTest3x3VoxelPoints_allwhite(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);

	expectedValue = 0;
	obtainedValue = nm.isExterior(v) ? 1 : 0;
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " Exterior Count Failed";

	runStandardTests(v, vpoints, nm, testName);
	testCount++;

	//////////////////////////////////////////////////////////////////////////
	//test exterior points - all black
	testName = "Test Exterior Points All Black - " + toString(testCount);
	vpoints.clear();
	generateTest3x3VoxelPoints_allblack(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);

	expectedValue = 1;
	obtainedValue = nm.isExterior(v) ? 1 : 0;
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " Exterior Count Failed";

	runStandardTests(v, vpoints, nm, testName);
	testCount++;


	//////////////////////////////////////////////////////////////////////////
	//test exterior points - case1
	testName = "Test Exterior Points Case 1- " + toString(testCount);
	vpoints.clear();
	generateTest3x3VoxelPoints_Case1(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);

	expectedValue = 1;
	obtainedValue = nm.isExterior(v)?1:0;
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " Exterior Count Failed";

	runStandardTests(v, vpoints, nm, testName);
	testCount++;
}



TEST(MorphologySkeletonizationTests, EndPointTest) {

	std::vector<int> vpoints;

	skeletonutil::Voxel v;
	skeletonutil::NAdjacencyMapper nm;

	auto expectedValue = 0;
	auto obtainedValue = 0;
	auto testCount = 0;

	v.pos_ = ivec3(1);
	v.index_ = nm.mapper3x3_(ivec3(1));
	v.distance_ = 1.0f;
	v.value_ = 1.0f;
	v.segmentationValue_ = 0;
	v.neighbours_ = 0;
	v.flags_ = 0;
	v.labels_ = 0;
	v.setVoxelFlag(skeletonutil::VoxelFlags::ForeGround);

	std::string testName;
	
	//////////////////////////////////////////////////////////////////////////
	//test end points - case1
	testName = "Test EndPoint Points Case 1 - " + toString(testCount);
	vpoints.clear();
	generateTest3x3VoxelPoints_Case0(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);

	expectedValue = 1;
	obtainedValue = nm.isEndPoint(v) ? 1 : 0;
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " EndPoint Count Failed";

	runStandardTests(v, vpoints, nm, testName);
	testCount++;

	//////////////////////////////////////////////////////////////////////////
	//test end points - case2
	testName = "Test EndPoint Points Case 2 - " + toString(testCount);
	vpoints.clear();
	generateTest3x3VoxelPoints_allblack(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);

	expectedValue = 1;
	obtainedValue = nm.isEndPoint(v) ? 1 : 0;
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " EndPoint Count Failed";

	runStandardTests(v, vpoints, nm, testName);
	testCount++;

	//////////////////////////////////////////////////////////////////////////
	//test end points - case2
	testName = "Test EndPoint Points Case 3 - " + toString(testCount);
	vpoints.clear();
	generateTest3x3VoxelPoints_Case1(vpoints, expectedValue);
	fillNeighbours(v, vpoints, nm);

	expectedValue = 0;
	obtainedValue = nm.isEndPoint(v) ? 1 : 0;
	EXPECT_EQ(expectedValue, obtainedValue) << testName << " EndPoint Count Failed";

	runStandardTests(v, vpoints, nm, testName);
	testCount++;
}


TEST(MorphologySkeletonizationTests, SegmentationTest) {

	size3_t dim(15);
	std::shared_ptr<Volume> volume = std::make_shared<Volume>(dim);
	auto binary = volume->getRepresentation<VolumeRAM>();
	

	std::function<bool(const size3_t&)> isForeGround = [&](const size3_t& pos) {
		if (pos.x>dim.x/2)
			return true;
		return false;
	};

	skeletonutil::VoxelMap allVoxels;

	util::IndexMapper3D index((dim));
	ivec3 rdim(ivec3(dim) - ivec3(1));
	K3DTree<size3_t, float> tree;	


	std::function<bool(size3_t)> voxelBoundaryTest = [&](size3_t newPos) {
		if (newPos.x < 0 || newPos.y < 0 || newPos.z < 0 || newPos.x > rdim.x || newPos.y > rdim.y ||
			newPos.z > rdim.z) {
			return false;
		}
		return true;
	};

	std::function<size_t(size3_t)> voxelMapping = [&](size3_t newPos) {
		return index(newPos);
	};

	std::function<vec3(size3_t)> voxelgetNormalizedPos = [&](size3_t newPos) {
		return vec3(newPos) / vec3(rdim);
	};

	std::function<std::vector<size3_t>(size3_t, double)> voxelNeighbourHood = [&](size3_t pos, double radius) {
		std::vector<size3_t> neighbours;
		auto nodes = tree.findCloseTo(pos, radius);
		for (auto& tnode : nodes) {
			auto p = tnode->get();
			auto indxp = voxelMapping(p);

			if (allVoxels.find(indxp)!=allVoxels.end() && 
				allVoxels.find(indxp)->second.hasVoxelFlag(VoxelFlags::ForeGround)) {
				neighbours.push_back(p);
			}
		}
		return neighbours;
	};

	//construct voxels
	auto func = [&](const size3_t& pos) {

		auto cpos = glm::clamp(ivec3(pos), ivec3(0), ivec3(dim) - ivec3(1));

		// Voxel flags: exterior, queued, deletable

		ivwAssert(allVoxels.find(index(cpos)) == allVoxels.end(), "Voxel entry already exist");

		//allVoxels[index(cpos)] = Voxel();
		//Voxel& v = allVoxels[index(cpos)];
		skeletonutil::Voxel v;
		v.pos_ = cpos;
		v.index_ = index(cpos);
		v.distance_ = (double(pos.x)/(dim.x/2.0));
		v.value_ = (pos.x > dim.x / 2) ? 1.0 : 0.0;
		v.neighbours_ = 0;
		v.flags_ = 0;
		v.labels_ = 0;
		v.segmentationValue_ = 0.0;
		v.mapping_ = voxelMapping;
		v.boundaryTest_ = voxelBoundaryTest;
		v.getNeighbours_ = voxelNeighbourHood;
		v.getNormalizedVoxelPos_ = voxelgetNormalizedPos;

		for (int x = 0; x < 3; x++)
			for (int y = 0; y < 3; y++)
				for (int z = 0; z < 3; z++)
					v.unsetNeighbour(ivec3(x, y, z));

		v.setVoxelFlag(VoxelFlags::None);

		if (isForeGround(cpos))
			v.setVoxelFlag(VoxelFlags::ForeGround);
		else
			v.setVoxelFlag(VoxelFlags::BackGround);


		allVoxels[index(cpos)] = v;

		tree.insert(cpos, cpos);

	};
	util::forEachVoxel(*binary, func);


	//extract foreground and background
	VoxelMap foregroundVoxels;
	VoxelMap backgroundVoxels;

	skeletonutil::getFlagedVoxels(allVoxels, foregroundVoxels, VoxelFlags::ForeGround); //foregrounds
	skeletonutil::getFlagedVoxels(allVoxels, backgroundVoxels, VoxelFlags::BackGround); //backgrounds

	//set neighbours
	for (auto& ov : allVoxels)
		skeletonutil::setNeighbourInfo(allVoxels, ov.second, dim);

	for (auto& fv : foregroundVoxels)
		skeletonutil::setNeighbourInfo(allVoxels, fv.second, dim);

	for (auto& bv : backgroundVoxels)
		skeletonutil::setNeighbourInfo(allVoxels, bv.second, dim);


	int testCount = 0;
    std::string testName;
	int expectedValue = 1;
	int obtainedValue = 0;

    //////////////////////////////////////////////////////////////////////////
    //test end points - case1
    testName = "Test Voxel Segmentation Case 1 - " + toString(testCount);
    
	//compute segments
    expectedValue = 1;
	auto segments = getLabeledVoxelSegments(foregroundVoxels, 0.0);

	obtainedValue = static_cast<int>(segments.size());
    EXPECT_EQ(expectedValue, obtainedValue) << testName << " Voxel Segmentation Failed";

    testCount++;
}


}