/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2017-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#ifndef IVW_GRADIENTSAMPLER_H
#define IVW_GRADIENTSAMPLER_H

#include <modules/morphologyvisualization/morphologyvisualizationmoduledefine.h>
#include <inviwo/core/common/inviwo.h>

#include <modules/morphologyvisualization/util/batchvolumesampler.h>

#include <vector>

namespace inviwo {

class Volume;

template <SampleBehaviour SB>
class TrajectoryVolumeSampler {
public:
    TrajectoryVolumeSampler(std::shared_ptr<const Volume> vol);
    ~TrajectoryVolumeSampler() = default;

    std::shared_ptr<const Volume> getVolume() const;

    dvec3 sample(const dvec3& worldPos, const dvec3& origin, const dvec3& extent);
    void sample(const std::vector<dvec3>& worldPos, std::vector<dvec3>& dest,
                const dvec3& origin, const dvec3& extent);

private:
    std::unique_ptr<BatchVolumeSampler<3>> volumeSampler_;
    std::shared_ptr<const Volume> volume_;

    std::vector<dvec3> samplePos_; //!< internal position cache
};


template <SampleBehaviour SB>
TrajectoryVolumeSampler<SB>::TrajectoryVolumeSampler(std::shared_ptr<const Volume> vol)
    : volumeSampler_(
        util::makeBatchVolumeSampler<3, SB, BoundaryBehaviour::Repeat,
        BoundaryBehaviour::Repeat, BoundaryBehaviour::Clamp>(vol))
    , volume_(vol) {}

template <SampleBehaviour SB>
std::shared_ptr<const Volume> TrajectoryVolumeSampler<SB>::getVolume() const { return volume_; }

template <SampleBehaviour SB>
dvec3 TrajectoryVolumeSampler<SB>::sample(const dvec3& worldPos, const dvec3& origin,
                                      const dvec3& extent) {
    if (!volume_) {
        return dvec3(0.0);
    }
    std::vector<dvec3> dest(1);
    sample({ worldPos }, dest, origin, extent);
    return dest.front();
}

template <SampleBehaviour SB>
void TrajectoryVolumeSampler<SB>::sample(const std::vector<dvec3>& worldPos, std::vector<dvec3>& dest,
                                     const dvec3& origin, const dvec3& extent) {
    if (!volume_) {
        std::fill(dest.begin(), dest.end(), dvec3(0.0));
        return;
    }

    samplePos_.resize(worldPos.size());
    // transform all input positions into texture coordinate space
    std::transform(worldPos.begin(), worldPos.end(), samplePos_.begin(), [origin, extent](const dvec3& p) {
        return ((p - origin) / extent);
    });

    volumeSampler_->sample(samplePos_, dest);
}



template <SampleBehaviour SB>
class ScalarVolumeSampler {
public:
    ScalarVolumeSampler(std::shared_ptr<const Volume> vol);
    ~ScalarVolumeSampler() = default;

    std::shared_ptr<const Volume> getVolume() const;

    double sample(const dvec3& worldPos, const dvec3& origin, const dvec3& extent);
    void sample(const std::vector<dvec3>& worldPos, std::vector<double>& dest,
                const dvec3& origin, const dvec3& extent);

private:
    std::unique_ptr<BatchVolumeSampler<1>> volumeSampler_;
    std::shared_ptr<const Volume> volume_;

    std::vector<dvec3> samplePos_; //!< internal position cache
};


template <SampleBehaviour SB>
ScalarVolumeSampler<SB>::ScalarVolumeSampler(std::shared_ptr<const Volume> vol)
    : volumeSampler_(
        util::makeBatchVolumeSampler<1, SB, BoundaryBehaviour::Repeat,
        BoundaryBehaviour::Repeat, BoundaryBehaviour::Clamp>(vol))
    , volume_(vol) {}

template <SampleBehaviour SB>
std::shared_ptr<const Volume> ScalarVolumeSampler<SB>::getVolume() const { return volume_; }

template <SampleBehaviour SB>
double ScalarVolumeSampler<SB>::sample(const dvec3& worldPos, const dvec3& origin,
                                          const dvec3& extent) {
    if (!volume_) {
        return 0.0;
    }
    std::vector<double> dest(1);
    sample({ worldPos }, dest, origin, extent);
    return dest.front();
}

template <SampleBehaviour SB>
void ScalarVolumeSampler<SB>::sample(const std::vector<dvec3>& worldPos, std::vector<double>& dest,
                                         const dvec3& origin, const dvec3& extent) {
    if (!volume_) {
        std::fill(dest.begin(), dest.end(), 0.0);
        return;
    }

    samplePos_.resize(worldPos.size());
    // transform all input positions into texture coordinate space
    std::transform(worldPos.begin(), worldPos.end(), samplePos_.begin(), [origin, extent](const dvec3& p) {
        return ((p - origin) / extent);
    });

    volumeSampler_->sample(samplePos_, dest);
}


} // namespace inviwo

#endif // IVW_GRADIENTSAMPLER_H
