/*********************************************************************************
 *
 * Inviwo - Interactive Visualization Workshop
 *
 * Copyright (c) 2016-2019 Inviwo Foundation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************************/

#ifndef IVW_RIDGEASSIGNMENT_H
#define IVW_RIDGEASSIGNMENT_H

#include <modules/morphologyvisualization/morphologyvisualizationmoduledefine.h>
#include <inviwo/core/common/inviwo.h>

namespace inviwo {

namespace morphviz {

template <typename InputIter, typename ValueType = typename std::iterator_traits<InputIter>::value_type>
std::vector<std::pair<ValueType, ValueType>> assignRidgesFromMaximas(const InputIter inputBegin, const InputIter inputEnd) {
    static_assert(
        std::is_base_of<std::bidirectional_iterator_tag,
                        typename std::iterator_traits<InputIter>::iterator_category>::value,
        "Iterators have to be bidirectional");
    static_assert(std::is_integral<typename std::iterator_traits<InputIter>::value_type>::value,
                  "Value type as to be integral");

    auto size = std::distance(inputBegin, inputEnd);
    if (size <= 1) return {};

    using Result = typename std::vector<std::pair<ValueType, ValueType>>;
    using OutIter = typename Result::iterator;

    Result output{ static_cast<size_t>(size), {0, 0} };

    const OutIter outputBegin = output.begin();
    const OutIter outputEnd = output.end();


    struct Iters {
        bool valid;
        char direction;
        InputIter in;
        OutIter out;
    };
    std::vector<Iters> iterators;

    auto toRidge = [](const ValueType& a, const ValueType& b) -> std::pair<ValueType, ValueType> {
        return{ std::min(a, b), std::max(a, b) };
    };

    auto inIt = inputBegin;
    auto outIt = outputBegin;
    for (; inIt != std::prev(inputEnd) && outIt != std::prev(outputEnd); ++inIt, ++outIt) {
        if (*inIt != *std::next(inIt)) {
            iterators.push_back({true, -1, inIt, outIt});
            iterators.push_back({true, 1, inIt, outIt});
            *outIt = toRidge(*inIt, *std::next(inIt));
        }
    }

    bool iterate = true;

    while (iterate) {
        iterate = false;
        for (auto iter = iterators.begin(); iter != iterators.end(); ++iter) {
            if (!iter->valid) continue;

            if ((iter->direction < 0 && iter->in != inputBegin) || (iter->direction > 0 && iter->in != std::prev(inputEnd))) {
                auto inNext = std::next(iter->in, iter->direction);
                auto outNext = std::next(iter->out, iter->direction);

                if (outNext->first == 0 && outNext->second == 0) {
                    if (*inNext == iter->out->first || *inNext == iter->out->second) {
                        *outNext = *iter->out;
                        iter->out = outNext;
                        iter->in = inNext;
                        iterate = true;
                    } else {    
                        iter->valid = false;
                    }
                } else {
                    iter->valid = false;
                }   
            }
        }
    }
    return output;
}

}  // namespace

}  // namespace

#endif // IVW_RIDGEASSIGNMENT_H

